/* $Id$

(c) Copyright, Real-Time Innovations, 2024.
All rights reserved.

No duplications, whole or partial, manual or electronic, may be made
without express written permission.  Any such copies, or
revisions thereof, must display this notice unaltered.
This code contains trade secrets of Real-Time Innovations, Inc.
*/

#ifndef RTI_CONFIG_COMPLIANCE_HPP_
#define RTI_CONFIG_COMPLIANCE_HPP_

// IMPORTANT: macros.hpp must be the first RTI header included in every header
// file so that symbols are exported correctly on Windows
#include <dds/core/macros.hpp>

#include <dds/core/types.hpp>

#include "ndds/ndds_config_c.h"

#include <bitset>

/**
 * \dref_GlobalConfigurationGroupDocs
 */

namespace rti { namespace config { namespace compliance {

/**
 * @headerfile rti/config/Compliance.hpp
 * @brief This mask configures different aspects of the Extended CDR encoding 
 * (XCDR and XCDR2).
 *
 * @details \dref_XTypesComplianceMask
 */
class OMG_DDS_API XTypesMask : public std::bitset<32> {
public:
    /**
     * @brief A typedef of std::bitset<32> for convenience
     */
    typedef std::bitset<32> MaskType;

public:
    /** \dref_XTypesComplianceMask_NDDS_CONFIG_XTYPES_COMPLIANCE_MASK_DEFAULT
    */
    XTypesMask()
    {
        MaskType(NDDS_CONFIG_XTYPES_COMPLIANCE_MASK_DEFAULT);
    }

    /**
     * @brief Construct a XTypesMask from an integer
     *
     * @param mask Value whose bits are copied to the bitset positions
     */
    explicit XTypesMask(uint64_t mask)
        : MaskType(mask)
    {
    }

    /**
     * @brief Construct a XTypesMask from a MaskType object
     *
     * @param mask A std::bitset<32> to construct this XTypesMask from
     */
    XTypesMask(const MaskType& mask)
        : MaskType(mask)
    {
    }

    /** \dref_XTypesComplianceMaskBits_NDDS_CONFIG_XTYPES_DHEADER_IN_NON_PRIMITIVE_COLLECTIONS_BIT
    */
    static const XTypesMask dheader_in_non_primitive_collections()
    {
        return XTypesMask(NDDS_CONFIG_XTYPES_DHEADER_IN_NON_PRIMITIVE_COLLECTIONS_BIT);
    }

    /** \dref_XTypesComplianceMaskBits_NDDS_CONFIG_XTYPES_ENUM_AS_PRIMITIVE_IN_COLLECTIONS_BIT
    */
    static const XTypesMask enum_as_primitive_in_collections()
    {
        return XTypesMask(NDDS_CONFIG_XTYPES_ENUM_AS_PRIMITIVE_IN_COLLECTIONS_BIT);
    }

    /** \dref_XTypesComplianceMaskBits_NDDS_CONFIG_XTYPES_PARAMETER_LENGTH_WITH_PADDING_BIT
    */
    static const XTypesMask parameter_length_with_padding()
    {
        return XTypesMask(NDDS_CONFIG_XTYPES_PARAMETER_LENGTH_WITH_PADDING_BIT);
    }

    /** \dref_XTypesComplianceMaskBits_NDDS_CONFIG_XTYPES_ENCAPSULATION_OPTIONS_WITH_PADDING_BIT
    */
    static const XTypesMask encapsulation_options_with_padding()
    {
        return XTypesMask(NDDS_CONFIG_XTYPES_ENCAPSULATION_OPTIONS_WITH_PADDING_BIT);
    }

    /** \dref_XTypesComplianceMask_NDDS_CONFIG_XTYPES_COMPLIANCE_MASK_VENDOR
    */
    static const XTypesMask vendor()
    {
        return XTypesMask(NDDS_CONFIG_XTYPES_COMPLIANCE_MASK_VENDOR);
    }

    /** \dref_XTypesComplianceMask_NDDS_CONFIG_XTYPES_COMPLIANCE_MASK_DEFAULT
    */
    static const XTypesMask default_mask()
    {
        return XTypesMask(NDDS_CONFIG_XTYPES_COMPLIANCE_MASK_DEFAULT);
    }
};

/**
 * @ingroup NDDSGlobalConfigurationModule
 * @brief Gets the XTypes compliance mask value for the application.
 *
 * \p \#include \p <rti/config/Compliance.hpp>
 *
 * @namespace_note rti::config::compliance
 *
 * @dref_details_XTypesCompliance_get_xtypes_compliance_mask
 */
OMG_DDS_API
XTypesMask get_xtypes_mask();

/**
 * @ingroup NDDSGlobalConfigurationModule
 * @brief Sets the level of compliance of the application with the
 * XTypes specification.
 *
 * \p \#include \p <rti/config/Compliance.hpp>
 *
 * @namespace_note rti::config::compliance
 *
 * @dref_details_XTypesCompliance_set_xtypes_compliance_mask
 */
OMG_DDS_API
void set_xtypes_mask(XTypesMask mask);


/**
 * @ingroup NDDSGlobalConfigurationModule
 * @brief Load the compliance masks from the environment variables.
 *
 * \p \#include \p <rti/config/Compliance.hpp>
 *
 * @namespace_note rti::config::compliance
 *
 * @dref_details_XTypesCompliance_load_compliance_masks
 */
OMG_DDS_API
void load_compliance_masks();

} } } // namespace rti::config::compliance

#endif /* RTI_CONFIG_COMPLIANCE_HPP_ */