/* $Id$

(c) Copyright, Real-Time Innovations, 2014-2016.
All rights reserved.

No duplications, whole or partial, manual or electronic, may be made
without express written permission.  Any such copies, or
revisions thereof, must display this notice unaltered.
This code contains trade secrets of Real-Time Innovations, Inc.
*/

#ifndef RTI_UTIL_LOGGER_HPP_
#define RTI_UTIL_LOGGER_HPP_

// IMPORTANT: macros.hpp must be the first RTI header included in every header
// file so that symbols are exported correctly on Windows
#include <dds/core/macros.hpp>

#include <functional>

#include "ndds/ndds_config_c.h"

#include <dds/core/LengthUnlimited.hpp>
#include <dds/core/SafeEnumeration.hpp>
#include <dds/core/Duration.hpp>
#include <rti/core/Exception.hpp>

/**
 * \dref_LoggingGroupDocs
 */

namespace rti { 

/**
 * @brief @extension Logging configuration and version information.
 */
namespace config {

/**
 * @ingroup DDSLoggingModule
 * @headerfile rti/config/Logger.hpp
 * @brief The definition of the dds::core::safe_enum LogLevel;
 */
struct LogLevel_def {

    typedef NDDS_Config_LogLevel native_type;

    /**
     * @brief The underlying \p enum type
     */
    enum type {
        /** \dref_LogLevel_NDDS_CONFIG_LOG_LEVEL_FATAL_ERROR */
        FATAL_ERROR = NDDS_CONFIG_LOG_LEVEL_FATAL_ERROR,

        /** \dref_LogLevel_NDDS_CONFIG_LOG_LEVEL_ERROR */
        EXCEPTION = NDDS_CONFIG_LOG_LEVEL_ERROR,

        /** \dref_LogLevel_NDDS_CONFIG_LOG_LEVEL_WARNING */
        WARNING = NDDS_CONFIG_LOG_LEVEL_WARNING,

        /** \dref_LogLevel_NDDS_CONFIG_LOG_LEVEL_STATUS_LOCAL */
        STATUS_LOCAL = NDDS_CONFIG_LOG_LEVEL_STATUS_LOCAL,

        /** \dref_LogLevel_NDDS_CONFIG_LOG_LEVEL_STATUS_REMOTE */
        STATUS_REMOTE = NDDS_CONFIG_LOG_LEVEL_STATUS_REMOTE,

        /** \dref_LogLevel_NDDS_CONFIG_LOG_LEVEL_DEBUG */
        STATUS_ALL = NDDS_CONFIG_LOG_LEVEL_DEBUG
    };
};

/**
 * @ingroup DDSLoggingModule
 * @brief The log level at which \ndds diagnostic information is logged.
 *
 * @see LogLevel_def
 */
typedef dds::core::safe_enum<LogLevel_def> LogLevel;


/** 
 * @ingroup DDSLoggingModule
 * @headerfile rti/config/Logger.hpp
 * @brief The definition of the dds::core::safe_enum Verbosity
 */
struct Verbosity_def {

    typedef NDDS_Config_LogVerbosity native_type;

    /**
     * @brief The underlying \p enum type
     */
    enum type {
        /** \dref_LogVerbosity_NDDS_CONFIG_LOG_VERBOSITY_SILENT */
        silent = RTI_LOG_BITMAP_NONE,

        /** \dref_LogVerbosity_NDDS_CONFIG_LOG_VERBOSITY_ERROR */
        exception = RTI_LOG_BIT_EXCEPTION | RTI_LOG_BIT_FATAL_ERROR,

        /** \dref_LogVerbosity_NDDS_CONFIG_LOG_VERBOSITY_WARNING */
        warning = exception | RTI_LOG_BIT_WARN,

        /** \dref_LogVerbosity_NDDS_CONFIG_LOG_VERBOSITY_STATUS_LOCAL */
        status_local = warning | RTI_LOG_BIT_LOCAL,

        /** \dref_LogVerbosity_NDDS_CONFIG_LOG_VERBOSITY_STATUS_REMOTE */
        status_remote = status_local | RTI_LOG_BIT_REMOTE,

        /** \dref_LogVerbosity_NDDS_CONFIG_LOG_VERBOSITY_STATUS_ALL */
        status_all = status_remote | RTI_LOG_BIT_PERIODIC,

        // for backward compatibility:
        SILENT = silent,
        EXCEPTION = exception,
        WARNING = warning,
        STATUS_LOCAL = status_local,
        STATUS_REMOTE = status_remote,
        STATUS_ALL = status_all,
        ERRORY = EXCEPTION
    };
};

/**
 * @ingroup DDSLoggingModule 
 * @brief The verbosities at which \ndds diagnostic information is logged. 
 *
 * @see Verbosity_def
 */
typedef dds::core::safe_enum<Verbosity_def> Verbosity;

/**
 * @ingroup DDSLoggingModule
 * @headerfile rti/config/Logger.hpp
 * @brief \dref_SyslogLevel
 */
enum class SyslogLevel {
    /** \dref_SyslogLevel_NDDS_CONFIG_SYSLOG_LEVEL_EMERGENCY */
    emergency = NDDS_CONFIG_SYSLOG_LEVEL_EMERGENCY,
    /** \dref_SyslogLevel_NDDS_CONFIG_SYSLOG_LEVEL_ALERT */
    alert = NDDS_CONFIG_SYSLOG_LEVEL_ALERT,
    /** \dref_SyslogLevel_NDDS_CONFIG_SYSLOG_LEVEL_CRITICAL */
    critical = NDDS_CONFIG_SYSLOG_LEVEL_CRITICAL,
    /** \dref_SyslogLevel_NDDS_CONFIG_SYSLOG_LEVEL_ERROR */
    error = NDDS_CONFIG_SYSLOG_LEVEL_ERROR,
    /** \dref_SyslogLevel_NDDS_CONFIG_SYSLOG_LEVEL_WARNING */
    warning = NDDS_CONFIG_SYSLOG_LEVEL_WARNING,
    /** \dref_SyslogLevel_NDDS_CONFIG_SYSLOG_LEVEL_NOTICE */
    notice = NDDS_CONFIG_SYSLOG_LEVEL_NOTICE,
    /** \dref_SyslogLevel_NDDS_CONFIG_SYSLOG_LEVEL_INFORMATIONAL */
    informational = NDDS_CONFIG_SYSLOG_LEVEL_INFORMATIONAL,
    /** \dref_SyslogLevel_NDDS_CONFIG_SYSLOG_LEVEL_DEBUG */
    debug = NDDS_CONFIG_SYSLOG_LEVEL_DEBUG
};

/**
 * @ingroup DDSLoggingModule
 * @headerfile rti/config/Logger.hpp
 * @brief \dref_SyslogVerbosity
 */
enum class SyslogVerbosity {
    /** \dref_SyslogVerbosity_NDDS_CONFIG_SYSLOG_VERBOSITY_SILENT */
    silent = NDDS_CONFIG_SYSLOG_VERBOSITY_SILENT,
    /** \dref_SyslogVerbosity_NDDS_CONFIG_SYSLOG_VERBOSITY_EMERGENCY */
    emergency = NDDS_CONFIG_SYSLOG_VERBOSITY_EMERGENCY,
    /** \dref_SyslogVerbosity_NDDS_CONFIG_SYSLOG_VERBOSITY_ALERT */
    alert = NDDS_CONFIG_SYSLOG_VERBOSITY_ALERT,
    /** \dref_SyslogVerbosity_NDDS_CONFIG_SYSLOG_VERBOSITY_CRITICAL */
    critical = NDDS_CONFIG_SYSLOG_VERBOSITY_CRITICAL,
    /** \dref_SyslogVerbosity_NDDS_CONFIG_SYSLOG_VERBOSITY_ERROR */
    error = NDDS_CONFIG_SYSLOG_VERBOSITY_ERROR,
    /** \dref_SyslogVerbosity_NDDS_CONFIG_SYSLOG_VERBOSITY_WARNING */
    warning = NDDS_CONFIG_SYSLOG_VERBOSITY_WARNING,
    /** \dref_SyslogVerbosity_NDDS_CONFIG_SYSLOG_VERBOSITY_NOTICE */
    notice = NDDS_CONFIG_SYSLOG_VERBOSITY_NOTICE,
    /** \dref_SyslogVerbosity_NDDS_CONFIG_SYSLOG_VERBOSITY_INFORMATIONAL */
    informational = NDDS_CONFIG_SYSLOG_VERBOSITY_INFORMATIONAL,
    /** \dref_SyslogVerbosity_NDDS_CONFIG_SYSLOG_VERBOSITY_DEBUG */
    debug = NDDS_CONFIG_SYSLOG_VERBOSITY_DEBUG
};

/**
 * @ingroup DDSLoggingModule
 * @headerfile rti/config/Logger.hpp
 * @brief The definition of the dds::core::safe_enum LogCategory
 */
struct LogCategory_def {

    typedef NDDS_Config_LogCategory native_type;

    /**
     * @brief The underlying \p enum type
     */
    enum type {
        /** \dref_LogCategory_NDDS_CONFIG_LOG_CATEGORY_PLATFORM
         */
        platform = NDDS_CONFIG_LOG_CATEGORY_PLATFORM,

        /** \dref_LogCategory_NDDS_CONFIG_LOG_CATEGORY_COMMUNICATION
         */
        communication = NDDS_CONFIG_LOG_CATEGORY_COMMUNICATION,

        /** \dref_LogCategory_NDDS_CONFIG_LOG_CATEGORY_DATABASE
         */
        database = NDDS_CONFIG_LOG_CATEGORY_DATABASE,

        /** \dref_LogCategory_NDDS_CONFIG_LOG_CATEGORY_ENTITIES
         */
        entities = NDDS_CONFIG_LOG_CATEGORY_ENTITIES,

        /** \dref_LogCategory_NDDS_CONFIG_LOG_CATEGORY_API
         */
        api = NDDS_CONFIG_LOG_CATEGORY_API,

        /** \dref_LogCategory_NDDS_CONFIG_LOG_CATEGORY_DISCOVERY
         */
        discovery = NDDS_CONFIG_LOG_CATEGORY_DISCOVERY,

        /** \dref_LogCategory_NDDS_CONFIG_LOG_CATEGORY_SECURITY
         */
        security = NDDS_CONFIG_LOG_CATEGORY_SECURITY,

        /** \dref_LogCategory_NDDS_CONFIG_LOG_CATEGORY_USER
         */
        user = NDDS_CONFIG_LOG_CATEGORY_USER,

        /** \dref_LogCategory_NDDS_CONFIG_LOG_CATEGORY_ALL
         */
        all_categories = NDDS_CONFIG_LOG_CATEGORY_ALL,

        // for backward compatibility:
  #ifndef PLATFORM // may be defined as a macro by other libraries
        PLATFORM = NDDS_CONFIG_LOG_CATEGORY_PLATFORM,
  #endif
        COMMUNICATION = NDDS_CONFIG_LOG_CATEGORY_COMMUNICATION,
        DATABASE = NDDS_CONFIG_LOG_CATEGORY_DATABASE,
        ENTITIES = NDDS_CONFIG_LOG_CATEGORY_ENTITIES,
        API = NDDS_CONFIG_LOG_CATEGORY_API,
        DISCOVERY = NDDS_CONFIG_LOG_CATEGORY_DISCOVERY,
        SECURITY = NDDS_CONFIG_LOG_CATEGORY_SECURITY,
        ALL_CATEGORIES = NDDS_CONFIG_LOG_CATEGORY_ALL
    };
};

/** 
 * @ingroup DDSLoggingModule 
 * @brief Categories of logged messages. 
 * @see @ref LogCategory_def   
 */
typedef dds::core::safe_enum<LogCategory_def> LogCategory;

/** 
 * @ingroup DDSLoggingModule
 * @headerfile rti/config/Logger.hpp
 * @brief The definition of the dds::core::safe_enum PrintFormat
 */
struct PrintFormat_def {

    typedef NDDS_Config_LogPrintFormat native_type;

    /**
     * @brief The underlying \p enum type
     */
    enum type {
        /*i \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_BIT_NUMBER
         */
        BIT_NUMBER = NDDS_CONFIG_LOG_PRINT_BIT_NUMBER,

        /*i \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_BIT_MESSAGE
         */
        BIT_MESSAGE = NDDS_CONFIG_LOG_PRINT_BIT_MESSAGE,

        /*i \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_BIT_LOCATION_MODULE
         */
        BIT_LOCATION_MODULE = NDDS_CONFIG_LOG_PRINT_BIT_LOCATION_MODULE,

        /*i \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_BIT_LOCATION_FILELINE
         */
        BIT_LOCATION_FILELINE = NDDS_CONFIG_LOG_PRINT_BIT_LOCATION_FILELINE,

        /*i \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_BIT_LOCATION_METHOD
         */
        BIT_LOCATION_METHOD = NDDS_CONFIG_LOG_PRINT_BIT_LOCATION_METHOD,

        /*i \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_BIT_TIMESTAMP
         */
        BIT_TIMESTAMP = NDDS_CONFIG_LOG_PRINT_BIT_TIMESTAMP,

        /*i \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_BIT_THREAD_ID
         */
        BIT_THREAD_ID = NDDS_CONFIG_LOG_PRINT_BIT_THREAD_ID,

        /*i \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_BIT_ACTIVITY_CONTEXT
         */
        BIT_ACTIVITY_CONTEXT = NDDS_CONFIG_LOG_PRINT_BIT_ACTIVITY_CONTEXT,

        /*i \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_BIT_TWO_LINES
         */
        BIT_TWO_LINES = NDDS_CONFIG_LOG_PRINT_BIT_TWO_LINES,

        /*i \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_BIT_BACKTRACE
         */
        BIT_BACKTRACE = NDDS_CONFIG_LOG_PRINT_BIT_BACKTRACE,

        /*i \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_BIT_LOG_LEVEL
         */
        BIT_LOG_LEVEL = NDDS_CONFIG_LOG_PRINT_BIT_LOG_LEVEL,

        /** \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_FORMAT_DEFAULT
         */
        DEFAULT = NDDS_CONFIG_LOG_PRINT_FORMAT_DEFAULT,

        /** \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_FORMAT_TIMESTAMPED
         */
        TIMESTAMPED = NDDS_CONFIG_LOG_PRINT_FORMAT_TIMESTAMPED,

        /** \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_FORMAT_VERBOSE
         */
        VERBOSE = NDDS_CONFIG_LOG_PRINT_FORMAT_VERBOSE,

        /** \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_FORMAT_VERBOSE_TIMESTAMPED
         */
        VERBOSE_TIMESTAMPED = NDDS_CONFIG_LOG_PRINT_FORMAT_VERBOSE_TIMESTAMPED,

        /** \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_FORMAT_DEBUG
         */
        DEBUG = NDDS_CONFIG_LOG_PRINT_FORMAT_DEBUG,

        /** \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_FORMAT_MINIMAL
         */
        MINIMAL = NDDS_CONFIG_LOG_PRINT_FORMAT_MINIMAL,

        /** \dref_LogPrintFormat_NDDS_CONFIG_LOG_PRINT_FORMAT_MAXIMAL
         */
        MAXIMAL = NDDS_CONFIG_LOG_PRINT_FORMAT_MAXIMAL

    };
};

/** 
 * @ingroup DDSLoggingModule 
 * @headerfile rti/config/Logger.hpp
 * @brief The format used to output RTI Connext diagnostic information.
 * @see @ref PrintFormat_def   
 */
typedef dds::core::safe_enum<PrintFormat_def> PrintFormat;

struct LoggerMode_def {

    typedef NDDS_Config_LoggerMode native_type;

    /**
     * @brief The underlying \p enum type
     */
    enum type {
        /*i \dref_LoggerMode_NDDS_CONFIG_LOGGER_MODE_DEFAULT,
         */
        DEFAULT = NDDS_CONFIG_LOGGER_MODE_DEFAULT,

        /*i \dref_LoggerMode_NDDS_CONFIG_LOGGER_MODE_DISABLE_STANDARD_DEVICES
         */
        DISABLE_STANDARD_DEVICES = NDDS_CONFIG_LOGGER_MODE_DISABLE_STANDARD_DEVICES,

        /*i \dref_LoggerMode_NDDS_CONFIG_LOGGER_MODE_DISABLE_ADVANCED_LOGGING
         */
        DISABLE_ADVANCED_LOGGING = NDDS_CONFIG_LOGGER_MODE_DISABLE_ADVANCED_LOGGING

    };
};

typedef dds::core::safe_enum<LoggerMode_def> LoggerMode;

/**
 * @ingroup DDSLoggingModule
 * @headerfile rti/config/Logger.hpp
 * @brief \dref_LogFacility
 */
enum class LogFacility {
    /**
     * \dref_LogFacility_NDDS_CONFIG_LOG_FACILITY_USER
     */
    user = NDDS_CONFIG_LOG_FACILITY_USER,
    /**
     * \dref_LogFacility_NDDS_CONFIG_LOG_FACILITY_SECURITY_EVENT
     */
    security_event = NDDS_CONFIG_LOG_FACILITY_SECURITY_EVENT,
    /**
     * \dref_LogFacility_NDDS_CONFIG_LOG_FACILITY_SERVICE
     */
    service = NDDS_CONFIG_LOG_FACILITY_SERVICE,
    /**
     * \dref_LogFacility_NDDS_CONFIG_LOG_FACILITY_MIDDLEWARE
     */
    middleware = NDDS_CONFIG_LOG_FACILITY_MIDDLEWARE
};

/**
 * @ingroup DDSLoggingModule
 * @headerfile rti/config/Logger.hpp
 * @brief A log message, including the text and additional information
 *
 * This struct passed to the handler set in Logger::output_handler().
 */
struct LogMessage {
    /**
     * @brief The log level of this message
     */
    LogLevel level;
    /**
     * @brief The text of this message.
     */
    const char *text;
    /**
     * @brief Indicates if the message is a security-related message.
     * @details \dref_details_LogMessage_is_security_message
     */
    bool is_security_message;
    /**
     * @brief A numeric code that identifies a specific log message.
     */
    unsigned int message_id;
    /**
     * @brief The time when the log message was printed.
     */
    dds::core::Duration timestamp;
    /**
     * @brief The Facility associated with the log message.
     * @see LogFacility
     */
    LogFacility facility;
};

/**
 * @ingroup DDSLoggingModule
 * @headerfile rti/config/Logger.hpp
 * @brief The singleton type used to configure RTI Connext logging.
 */
class OMG_DDS_API Logger {
public:
    /** \dref_Logger_get_instance
     */
    static Logger& instance()
    {
        static Logger singleton;
        return singleton;
    }

    /** \dref_Logger_get_verbosity
     */
    Verbosity verbosity();

    /** \dref_Logger_get_verbosity_by_category
     */
    Verbosity verbosity_by_category(LogCategory category);

    /** \dref_Logger_set_verbosity
     */
    void verbosity(Verbosity verbosity);

    /** \dref_Logger_set_verbosity_by_category
     */
    void verbosity_by_category(LogCategory category, Verbosity verbosity);

    /** \dref_Logger_get_output_file
     */
    FILE* output_file();

    /** \dref_Logger_set_output_file
     */
    void output_file(FILE* out);

    /** \dref_Logger_set_output_file_name
     */
    void output_file(const char * file_name);

    /** \dref_Logger_set_output_file_set
     */
    void output_file_set(
            const char * file_preffix,
            const char * file_suffix,
            int max_bytes,
            int max_files = dds::core::LENGTH_UNLIMITED);

    /**
     * @brief Redirects RTI Connext logging to a function
     *
     * When a handler is set, log messages stop being directed to the standard
     * output and are passed to the handler.
     *
     * To direct logging back to the standard output, use
     * reset_output_handler().
     *
     * @param handler A function or function object capable of receiving a
     * `const LogMessage&` argument, the log message.
     *
     * The following example sets a handler that saves all log messages in
     * a vector:
     * ```
     * std::vector<std::string> saved_logs;
     * Logger::instance().output_handler([&saved_logs](const LogMessage& message) {
     *     saved_logs.push_back(message.text); // message is copied as a std::string
     * });
     * ```
     *
     * The `char *` in the `LogMessage` is only valid within the callback.
     */
    void output_handler(std::function<void(const LogMessage&)> handler)
    {
        reset_output_handler();
        if (handler) {
            handler_ = std::move(handler);
            set_device();
        }
    }

    /**
     * @brief Resets a previously set output handler
     *
     * Directs RTI Connext logging back to the standard output.
     */
    void reset_output_handler()
    {
        NDDS_Config_Logger_set_output_device(native_logger_, NULL);
        handler_ = std::function<void(const LogMessage &)>();
    }

    /** \dref_Logger_get_print_format
     */
    PrintFormat print_format();

    /** \dref_Logger_set_print_format
     */
    void print_format(PrintFormat print_format);

    /** \dref_Logger_get_print_format_by_log_level
     */
    PrintFormat print_format_by_log_level(LogLevel log_level);

    /** \dref_Logger_set_print_format_by_log_level
     */
    void print_format_by_log_level(PrintFormat print_format, LogLevel log_level);

    /*i \dref_Logger_set_mode
     */
    void mode(LoggerMode mode);

    /** \dref_Logger_emergency
     */
    void emergency(const char *msg);

    /** \dref_Logger_alert
     */
    void alert(const char *msg);

    /** \dref_Logger_critical
     */
    void critical(const char *msg);

    /** \dref_Logger_error
     */
    void error(const char *msg);

    /** \dref_Logger_warning
     */
    void warning(const char *msg);

    /** \dref_Logger_notice
     */
    void notice(const char *msg);

    /** \dref_Logger_informational
     */
    void informational(const char *msg);

    /** \dref_Logger_debug
     */
    void debug(const char *msg);

#if defined(__cpp_lib_string_view) || defined(DOXYGEN_DOCUMENTATION_ONLY)
    /** \dref_Logger_emergency
     */
    void emergency(std::string_view msg);

    /** \dref_Logger_alert
     */
    void alert(std::string_view msg);

    /** \dref_Logger_critical
     */
    void critical(std::string_view msg);

    /** \dref_Logger_error
     */
    void error(std::string_view msg);

    /** \dref_Logger_warning
     */
    void warning(std::string_view msg);

    /** \dref_Logger_notice
     */
    void notice(std::string_view msg);

    /** \dref_Logger_informational
     */
    void informational(std::string_view msg);

    /** \dref_Logger_debug
     */
    void debug(std::string_view msg);
#endif

private:
    Logger() : native_logger_(NDDS_Config_Logger_get_instance())
    {
        device_.device_data = static_cast<void *>(this);
        device_.write = device_callback;
    }

    // Disable copy
    Logger(const Logger&);
    Logger& operator=(const Logger&);

    static void device_callback(
        struct NDDS_Config_LoggerDevice *device,
        const struct NDDS_Config_LogMessage *message)
    {
        Logger *this_logger = static_cast<Logger *>(device->device_data);

        if (this_logger != nullptr && message != nullptr) {
            try {
                if (this_logger->handler_) {
                    LogMessage msg;
                    msg.level = (LogLevel::type) message->level;
                    msg.text = message->text;
                    msg.is_security_message =
                            message->is_security_message == DDS_BOOLEAN_TRUE;
                    msg.message_id = message->message_id;
                    // TRUST-1652
                    // Intentional: Both enums refer to syslog facilities. Also,
                    // the enum values in C++ are defined using the C values,
                    // so the assignment is safe.
                    /* coverity[mixed_enums] */
                    msg.facility = (LogFacility) message->facility;
                    msg.timestamp.sec(message->timestamp.sec);
                    msg.timestamp.nanosec(message->timestamp.nanosec);
                    this_logger->handler_(msg);
                }
            } catch (...) {
                // suppress exceptions in C callback
            }
        }
    }

    void set_device()
    {
        DDS_ReturnCode_t retcode =
                NDDS_Config_Logger_set_output_device(native_logger_, &device_)
                ? DDS_RETCODE_OK
                : DDS_RETCODE_ERROR;
        if (retcode != DDS_RETCODE_OK) {
            reset_output_handler();
        }

        rti::core::check_return_code(retcode, "set output device");
    }

    NDDS_Config_Logger *native_logger_;  // implementation
    std::function<void(const LogMessage&)> handler_;
    NDDS_Config_LoggerDevice device_;
};

/** 
 * @ingroup DDSLoggingModule 
 * @headerfile rti/config/Logger.hpp
 * @brief Changes the logger verbosity temporarily during the scope of a variable.
 *
 * RAII utility to modify logging verbosity during the scope of a variable and
 * restore it to the previous value upon destruction.
 */
class ScopedLoggerVerbosity {
public:
    /**
     * Sets the verbosity of Logger::instance()
     */
    ScopedLoggerVerbosity(Verbosity verbosity)
        : previous_verbosity_ (Logger::instance().verbosity())
    {
        Logger::instance().verbosity(verbosity);
    }

    /**
     * Restores the verbosity of Logger::instance() to the value before the
     * construction of this object.
     */
    ~ScopedLoggerVerbosity()
    {
        Logger::instance().verbosity(previous_verbosity_);
    }

private:
    Verbosity previous_verbosity_;
};

} } // namespace rti::config

#endif // RTI_UTIL_LOGGER_HPP_
