/* $Id$

(c) Copyright, Real-Time Innovations, 2013-2016.
All rights reserved.

No duplications, whole or partial, manual or electronic, may be made
without express written permission.  Any such copies, or
revisions thereof, must display this notice unaltered.
This code contains trade secrets of Real-Time Innovations, Inc.


modification history
--------------------
1.0a,22jan13,acr Created
============================================================================= */

#ifndef RTI_DDS_CORE_ENTITY_HPP_
#define RTI_DDS_CORE_ENTITY_HPP_

#include <string>
#include <utility> // for std::pair
#include <functional> // for std::function

// IMPORTANT: macros.hpp must be the first RTI header included in every header
// file so that symbols are exported correctly on Windows
#include <dds/core/macros.hpp>

#include <dds/core/ref_traits.hpp>
#include <dds/core/status/Status.hpp>
#include <rti/core/detail/SelfReference.hpp>
#include <rti/core/cond/StatusConditionImpl.hpp>
#include <rti/core/detail/ListenerHolder.hpp>

namespace rti { namespace core {

class EntityLock;

class OMG_DDS_API Entity : public detail::RetainableType<Entity, 2> {
public:
    typedef detail::RetainableType<Entity, 2> RetainableType;
    typedef RetainableType::ref_type ref_type;
    typedef RetainableType::weak_ref_type weak_ref_type;

    virtual ~Entity();

protected:
    Entity(DDS_Entity *native_entity, detail::ListenerHolderPtr listener_holder);
    Entity(const std::pair<DDS_Entity *, bool> &native_entity_enable,
           detail::ListenerHolderPtr listener_holder);

private: // Disable copy
    Entity(const Entity&);
    Entity& operator=(const Entity& other);

public:
    void enable();
    const dds::core::status::StatusMask status_changes();
    const dds::core::InstanceHandle instance_handle() const;


    /* To be specialized by subclasses. The overridden method must call
     * Entity::close() at the end
     */
    virtual void close();
    void retain();

public:

    // --- RTI extensions: ----------------------------------------------------

    bool enabled() const;
    bool closed() const
    {
        return native_entity_ == NULL;
    }

public:
    // --- For internal use only: ---------------------------------------------

    void remember_reference(ref_type reference);

    DDS_Entity * native_entity()
    {
        assert_not_closed();
        return native_entity_;
    }

    const DDS_Entity * native_entity() const
    {
        assert_not_closed();
        return native_entity_;
    }

    DDS_Entity *native_entity_noexcept() noexcept
    {
        return native_entity_;
    }

    const DDS_Entity *native_entity_noexcept() const noexcept
    {
        return native_entity_;
    }

    void native_entity(DDS_Entity * entity);

    dds::core::smart_ptr_traits<rti::core::cond::StatusCondition>::ref_type
    status_condition_ref() const;

protected:
    void assert_not_closed() const
    {
        if (closed()) {
            throw dds::core::AlreadyClosedError("already closed");
        }
    }

    // must be called by subclasses when a listener is set
    void assert_legal_listener_setter_call(bool user_call) const;

    // must be called by subclasses when a listener is set
    void retain_for_listener();

    // must be called by subclasses the listener is unset
    void unretain_for_listener();

    virtual void reserved_data(void * ptr);
    virtual void * reserved_data();

public:
    void created_from_c(bool the_created_from_c)
    {
        created_from_c_ = the_created_from_c;
    }

    bool created_from_c() const
    {
        return created_from_c_;
    }

    bool is_user_created() const
    {
        return DDS_Entity_is_user_created(native_entity_noexcept());
    }

    void set_user_data_(void * user_data)
    {
        user_data_ = user_data;
    }

    void set_user_data_(void * user_data, std::function<void(void*)> deleter)
    {
        user_data_ = user_data;
        user_data_deleter_ = deleter;
    }

    void * get_user_data_() const
    {
        return user_data_;
    }

private:
    friend class EntityLock; // allow access to get_self_reference()

    static volatile unsigned int entityID_;
    DDS_Entity * native_entity_;
    mutable rti::core::cond::StatusCondition::weak_ref_type status_condition_weak_ref_;
    bool created_from_c_;
    bool needs_enable_;

    // stores an arbitrary object set by a higher layer
    void *user_data_ = nullptr;

    // provides a way to delete the user data right before the Entity is destroyed
    std::function<void(void*)> user_data_deleter_;
protected:
    detail::ListenerHolderPtr listener_holder_;
};

} } // namespace rti::core

#endif // RTI_DDS_CORE_ENTITY_HPP_

