/* $Id$

(c) Copyright, Real-Time Innovations, 2013-2016.
All rights reserved.

No duplications, whole or partial, manual or electronic, may be made
without express written permission.  Any such copies, or
revisions thereof, must display this notice unaltered.
This code contains trade secrets of Real-Time Innovations, Inc.
*/

#ifndef RTI_DDS_CORE_RESULT_HPP_
#define RTI_DDS_CORE_RESULT_HPP_

// IMPORTANT: macros.hpp must be the first RTI header included in every header
// file so that symbols are exported correctly on Windows
#include <dds/core/macros.hpp>

#include <rti/core/Exception.hpp>

#include "ndds/ndds_c.h"

namespace rti { namespace core {

/**
 * @ingroup DDSException
 * @brief A result from an operation that doesn't throw exceptions containing a
 * return code and (if successful) a value
 *
 * For some critical operations in the API a exception-throwing and a noexcept
 * version of the same function are provided. The noexcept versions return a
 * Result<T> object to indicate if the operation was successful or not and, if
 * it was, provide the return value of type T.
 *
 * Note that `Result<void>` is specialized to not provide a value, only a
 * return code.
 */
template <typename T>
class Result {
public:
    /**
     * @brief Creates a Result object with a return code of DDS_RETCODE_ERROR and no
     * value.
     */
    Result() = default;

    /**
     * @brief Creates a Result object with a return code of DDS_RETCODE_OK and
     * a value that is copied.
     */
    explicit Result(const T& value) : value_(value), retcode_(DDS_RETCODE_OK)
    { }

    /**
     * @brief Creates a Result object with a return code of DDS_RETCODE_OK and a
     * value that is moved.
     */
    explicit Result(T&& value) : value_(std::move(value)), retcode_(DDS_RETCODE_OK)
    { }

    /**
     * @brief Creates a Result object with a return code and no value.
     */
    explicit Result(DDS_ReturnCode_t retcode) : retcode_(retcode) { }

    // default copy and move
    Result(const Result& other) = default;
    Result(Result&& other) = default;
    Result& operator=(const Result& other) = default;
    Result& operator=(Result&& other) = default;

    /**
     * @brief Gets the value without checking if the operation was successful.
     *
     * If the operation was not successful, the behavior is undefined; use
     * is_ok() first or use get_if_ok() instead.
     */
    const T& get() const& noexcept
    {
        return value_;
    }

    /**
     * @brief Gets the value without checking if the operation was successful.
     *
     * If the operation was not successful, the behavior is undefined; use
     * is_ok() first or use get_if_ok() instead.
     */
    T& get() & noexcept
    {
        return value_;
    }

    // It's not safe to get the result without checking if there was an error;
    // to prevent getting the result from a temporary object, we delete this
    // overload; note that get_if_ok() can be safely used on a temporary object.
    T&& get() && = delete;

    /**
     * @brief Gets the return code.
     */
    DDS_ReturnCode_t get_return_code() const noexcept
    {
        return retcode_;
    }

    /**
     * @brief Returns true if the operation was successful
     *
     * This is equivalent to `get_return_code() == DDS_RETCODE_OK`
     */
    bool is_ok() const noexcept
    {
        return retcode_ == DDS_RETCODE_OK;
    }

    /**
     * @brief Gets the value if the operation was successful, otherwise throws
     * an exception.
     */
    const T& get_if_ok() const&
    {
        throw_if_error();
        return value_;
    }

    /**
     * @brief Gets the value if the operation was successful, otherwise throws
     * an exception.
     */
    T& get_if_ok() &
    {
        throw_if_error();
        return value_;
    }

    /**
     * @brief Gets the value if the operation was successful, otherwise throws
     * an exception.
     */
    T&& get_if_ok() &&
    {
        throw_if_error();
        return std::move(value_);
    }

    /**
     * @brief Throws an exception if the operation was not successful.
     *
     * If the operation was successful, this function does nothing.
     */
    void throw_if_error() const
    {
        rti::core::check_return_code(retcode_, "Result has an error");
    }

private:
    T value_ {};
    DDS_ReturnCode_t retcode_ {DDS_RETCODE_ERROR};
};

template <>
class Result<void> {
public:
    Result() = default;

    explicit Result(DDS_ReturnCode_t retcode) : retcode_(retcode)
    { }

    // default copy and move
    Result(const Result& other) = default;
    Result(Result&& other) = default;
    Result& operator=(const Result& other) = default;
    Result& operator=(Result&& other) = default;

    DDS_ReturnCode_t get_return_code() const noexcept
    {
        return retcode_;
    }

    bool is_ok() const noexcept
    {
        return retcode_ == DDS_RETCODE_OK;
    }

    void throw_if_error() const
    {
        rti::core::check_return_code(retcode_, "Result has an error");
    }

private:
    DDS_ReturnCode_t retcode_ {DDS_RETCODE_ERROR};
};

} }
#endif
