/* $Id$

 (c) Copyright, Real-Time Innovations, 2014-2016. 
 All rights reserved.
 No duplications, whole or partial, manual or electronic, may be made
 without express written permission.  Any such copies, or
 revisions thereof, must display this notice unaltered.
 This code contains trade secrets of Real-Time Innovations, Inc.

modification history
-------------------- 
5.0.2,12jun14,eam Created
============================================================================= */

/*i @file
   Defines utilities that can be used in the testers
*/
#ifndef RTI_TEST_UTILITIES_HPP_
#define RTI_TEST_UTILITIES_HPP_

#include <mutex>
#include <list>

#include <dds/core/QosProvider.hpp>
#include <dds/domain/DomainParticipant.hpp>
#include <dds/sub/DataReader.hpp>
#include <dds/pub/DataWriter.hpp>
#include <dds/sub/discovery.hpp>
#include <dds/pub/discovery.hpp>

#include <rti/config/Logger.hpp>

namespace rti { namespace test {

//
// Base class for defining threads by implementing the run() method with
// auto-join capability.
//
Package
class Thread {
private:
    std::string _name;
    bool _auto_join;
    bool _running;
    RTIOsapiThread * _thread;
    RTIOsapiSemaphore * _exit_sem;
    bool _exception;
    std::string _exception_message;
public:

    Thread(const std::string& name = std::string("Thread"),
              bool auto_join = true)
        : _name(name),
          _auto_join(auto_join),
          _running(false),
          _thread(NULL),
          _exit_sem(NULL),
          _exception(false)
    {
        RTIOsapiSemaphoreProperty sem_prop =
                RTI_OSAPI_SEMAPHORE_PROPERTY_DEFAULT;
        _exit_sem = RTIOsapiSemaphore_new(
                RTI_OSAPI_SEMAPHORE_KIND_BINARY, &sem_prop);
        RTI_TEST_ASSERT(_exit_sem != NULL);
    }

    // disable copies
    Thread(const Thread&) = delete;
    Thread& operator=(const Thread&) = delete;

    virtual ~Thread()
    {
        if (_auto_join && _running) {
            try {
                join();
            } catch(...) {
                return;
            }
        }
        RTIOsapiSemaphore_delete(_exit_sem); // Can't be NULL
        if(_thread != NULL) {
            RTIOsapiThread_delete(_thread);
        }
    }

    void join(int max_wait_sec = 10) {

        if (_running) {
            RTINtpTime  max_wait = {0, 0};
            max_wait.sec = max_wait_sec;
            if (RTIOsapiSemaphore_take(_exit_sem, &max_wait) !=
                    RTI_OSAPI_SEMAPHORE_STATUS_OK) {
                RTI_TEST_FAIL(_name + " join timed out");
            }
        }

        if (_exception) {
            RTI_TEST_FAIL(
                ("Thread finished with exception: " + _exception_message).c_str());
        }
    }

    void start() {
        _running = true;
        _thread = RTIOsapiThread_new(
            "TestUtilsThread",
            RTI_OSAPI_THREAD_PRIORITY_NORMAL,
            RTI_OSAPI_THREAD_OPTION_DEFAULT,
            RTI_OSAPI_THREAD_STACK_SIZE_DEFAULT, NULL,
            run_wrapper, this);

        if (_thread == NULL) {
            _running = false;
            RTI_TEST_FAIL("Failed to create RTIOsapiThread");
        }
    }

    virtual void run() = 0;

private:
    static void * run_wrapper(void * args)
    {
        const char * METHOD_NAME =
            "Thread::run_wrapper";
        Thread * thread = (Thread *) args;
        thread->_exception = false;
        try {
            thread->run();
        } catch (const std::exception& ex) {
            thread->_exception = true;
            thread->_exception_message = ex.what();
            RTITestLog_exception(&RTI_LOG_ANY_FAILURE_s,
                                 ex.what());
        } catch (...) {
            thread->_exception = true;
            thread->_exception_message = "Unknown exception";
            RTITestLog_exception(&RTI_LOG_ANY_FAILURE_s,
                                "Unknown exception");
        }

        thread->_running = false;
        RTIOsapiSemaphore_give(thread->_exit_sem);
        return NULL;
    }

};


// --- AutoEnableCreatedEntitiesGuard: ----------------------------------------

/*
 * A guard which sets the DPF's EntityFactoryQos false on creation and true on 
 * deletion 
 */
class AutoEnableCreatedEntitiesGuard {

public:
    AutoEnableCreatedEntitiesGuard()
    {
        // Create participants disabled 
        dds::domain::qos::DomainParticipantFactoryQos qos = 
            dds::domain::DomainParticipant::participant_factory_qos();
        qos << dds::core::policy::EntityFactory(false);
        dds::domain::DomainParticipant::participant_factory_qos(qos);
    }

    AutoEnableCreatedEntitiesGuard(
        dds::domain::qos::DomainParticipantFactoryQos& qos)
    {
        // Create participants disabled 
        qos << dds::core::policy::EntityFactory(false);
        dds::domain::DomainParticipant::participant_factory_qos(qos);
    }

    ~AutoEnableCreatedEntitiesGuard()
    {
        // Reset the DPF qos to create entities enabled for the next tests 
        dds::domain::qos::DomainParticipantFactoryQos qos  = 
            dds::domain::DomainParticipant::participant_factory_qos();
        qos << dds::core::policy::EntityFactory(true);
        dds::domain::DomainParticipant::participant_factory_qos(qos);
    }
}; 

// --- AutoEnableCreatedEntitiesGuard: ----------------------------------------

/*
 * A guard which resets the default QosProvider in case a unit test changes 
 * the default library or profile, those changes will not affect following tests 
 */
class DefaultQosProviderGuard {

public:
    DefaultQosProviderGuard() { }

    ~DefaultQosProviderGuard()
    {
        // Reset the Default QosProvider 
        dds::core::QosProvider::reset_default();
    }
};

// --- EnvVarToken: ----------------------------------------------------------

/**
 * This class is used to create or modify an environment variable.
 * If the variable didn't exist previously, the environment variable will be set
 * and it will last as long as the life of this instance.
 * If the variable already existed, this class will change its value to the
 * provided one during the lifetime of this instance. After destruction, it will
 * return to its previous value.
 */
class EnvVarToken {
public:
    EnvVarToken() {}
    EnvVarToken(const std::string& env_var_name, const std::string& value) :
        env_var_name_(env_var_name)
    {
        // Consider that a variable can hold the value of a path
        char current_env_var[RTI_OSAPI_STRING_SEQ_STRING_MAX_SIZE] = "";
        if (RTIOsapi_envVarOrFileGet(
                current_env_var,
                RTI_OSAPI_STRING_SEQ_STRING_MAX_SIZE,
                env_var_name.c_str(),
                NULL,
                // The following two not used with environment variables
                '.',
                '.') != NULL) {
            previous_value_.assign(current_env_var);
        }
        std::stringstream stream;
        stream << env_var_name << "=" << value;
        set_var(stream.str());
    }
    EnvVarToken(const std::string& env_var_name, const uint32_t value) :
        env_var_name_(env_var_name)
    {
        // Consider that a variable can hold the value of a path
        char current_env_var[RTI_OSAPI_STRING_SEQ_STRING_MAX_SIZE] = "";
        if (RTIOsapi_envVarOrFileGet(
                current_env_var,
                RTI_OSAPI_STRING_SEQ_STRING_MAX_SIZE,
                env_var_name.c_str(),
                NULL,
                // The following two not used with environment variables
                '.',
                '.') != NULL) {
            previous_value_.assign(current_env_var);
        }
        std::stringstream stream;
        stream << env_var_name << "=" << value;
        set_var(stream.str());
    }

    ~EnvVarToken()
    {
        const char *METHOD_NAME = RTI_FUNCTION_NAME;
        if (previous_value_.empty()) {
            if (!RTIOsapi_envVarUnset(env_var_name_.c_str())) {
                RTITestLog_exception(
                        &RTI_LOG_ANY_FAILURE_ss,
                        "Failed to unset environment variable: ",
                        env_var_name_.c_str());
            }
        } else {
            std::stringstream stream;
            stream << env_var_name_ << "=" << previous_value_;
            set_var(stream.str());
        }
    }

    void set(const std::string& env_var_name, const std::string& value)
    {
        env_var_name_ = env_var_name;
        std::stringstream stream;
        stream << env_var_name << "=" << value;
        set_var(stream.str());
    }

private:
    std::string env_var_name_;
    std::string previous_value_;
    /* 
     * Need to retain the memory of the parameter passed to RTIOsapi_envVarSet.
     * RTIOsapi_envVarSet could use putenv (for e.g.in case of QNX), where the 
     * call doesn't make a copy of the passed in value unlike setenv.
     */
    std::string env_var_name_value_pair_;

    void set_var(const std::string& set_var_expression)
    {
        env_var_name_value_pair_ = set_var_expression;
        if (!RTIOsapi_envVarSet(env_var_name_value_pair_.c_str())) {
            RTI_TEST_FAIL(
                    ("Failed to set up environment variable: " +
                    set_var_expression).c_str());
        }

    }
};

template<typename T>
void wait_for_discovery(
        dds::pub::DataWriter<T> writer,
        dds::sub::DataReader<T> reader)
{
    RTI_TEST_WAIT_FOR_EQUALS(
            1,
            dds::sub::matched_publications(reader).size(),
            RTI_TEST_DEFAULT_WAIT_TIME);
    RTI_TEST_WAIT_FOR_EQUALS(
            1,
            dds::pub::matched_subscriptions(writer).size(),
            RTI_TEST_DEFAULT_WAIT_TIME);
}

template<typename T>
void wait_for_data(dds::sub::DataReader<T> reader, int sample_count)
{
    RTI_TEST_WAIT_FOR_EQUALS(
            sample_count,
            static_cast<int>(reader.read().length()),
            RTI_TEST_DEFAULT_WAIT_TIME);
}

class LogCapture {
public:
    LogCapture()
    {
        rti::config::Logger::instance().output_handler(
                [&](const rti::config::LogMessage &message) {
                    std::lock_guard<std::mutex> lock(logger_mutex_);
                    saved_logs_ += message.text;
                });

        for (int i = 0; i < NDDS_CONFIG_COUNT_LOG_CATEGORIES; i++) {
            rti::config::LogCategory category =
                    static_cast<rti::config::LogCategory::inner_enum>(i);
            cached_verbosity_level_per_category_[i] =
                    rti::config::Logger::instance().verbosity_by_category(
                            category);
        }

    }

    LogCapture(const LogCapture&) = delete;
    LogCapture& operator=(const LogCapture&) = delete;

    ~LogCapture()
    {
        reset();
    }

    void reset()
    {
        rti::config::Logger::instance().reset_output_handler();

        for (int i = 0; i < NDDS_CONFIG_COUNT_LOG_CATEGORIES; i++) {
            rti::config::LogCategory category =
                    static_cast<rti::config::LogCategory::inner_enum>(i);
            rti::config::Logger::instance().verbosity_by_category(
                    category,
                    cached_verbosity_level_per_category_[i]);
        }
    }

    void clear()
    {
        std::lock_guard<std::mutex> lock(logger_mutex_);
        saved_logs_.clear();
    }

    bool empty()
    {
        std::lock_guard<std::mutex> lock(logger_mutex_);
        return saved_logs_.empty();
    }

    bool contains(const std::string &text)
    {
        return saved_logs_.find(text) != std::string::npos;
    }

private:
    std::mutex logger_mutex_;
    std::string saved_logs_;
    std::array<rti::config::Verbosity, NDDS_CONFIG_COUNT_LOG_CATEGORIES>
            cached_verbosity_level_per_category_;
};

template <typename TBase, class T>
void test_entity_in_containers(T& object1, T& object2)
{
    // The following assertion is a precondition for the test to work
    RTI_TEST_ASSERT(object1 != object2);

    const long initial_use_count1 = object1.delegate().use_count();
    const long initial_use_count2 = object2.delegate().use_count();

    std::vector<T> vector(5, object1);
    RTI_TEST_ASSERT(object1.delegate().use_count() == initial_use_count1 + 5);

    vector.push_back(object2);
    RTI_TEST_ASSERT(object2.delegate().use_count() == initial_use_count2 + 1);

    RTI_TEST_ASSERT(vector[3] == object1);
    RTI_TEST_ASSERT(vector[5] == object2);

    std::list<TBase> list(vector.begin(), vector.end());
    RTI_TEST_ASSERT(object1.delegate().use_count() == initial_use_count1 + 10);
    RTI_TEST_ASSERT(object2.delegate().use_count() == initial_use_count2 + 2);

    list.clear();
    RTI_TEST_ASSERT(object1.delegate().use_count() == initial_use_count1 + 5);
    RTI_TEST_ASSERT(object2.delegate().use_count() == initial_use_count2 + 1);
}

template <class ReferenceType>
void test_entity_swap(ReferenceType object1, ReferenceType object2)
{
    typedef typename ReferenceType::DELEGATE_T Delegate;

    RTI_TEST_ASSERT(object1 != object2); // test precondition

    Delegate* object1_ptr = object1.delegate().get();
    Delegate* object2_ptr = object2.delegate().get();

    swap(object1, object2);

    RTI_TEST_ASSERT(object1_ptr == object2.delegate().get());
    RTI_TEST_ASSERT(object2_ptr == object1.delegate().get());
}

/*
 * Test the reference semantics of entity classes (T is a class derived
 * from dds::core::Entity). Test that assignment to the base class is possible.
 *
 * @pre There are no other references to object1 and object2
 */
template <class Type, class BaseType, class NativeTypeGetter>
void test_reference_assignment(
    Type& object1,
    Type& object2,
    NativeTypeGetter native_getter = NativeTypeGetter())
{
    // The following three assertions are preconditions for the test to work
    RTI_TEST_ASSERT(object1.delegate().use_count() == 1);
    RTI_TEST_ASSERT(object2.delegate().use_count() == 1);
    RTI_TEST_ASSERT(object1 != object2);

    Type object3 = object1;
  #ifdef RTI_NO_CXX11_NULLPTR
    Type object4(dds::core::null);
  #else
    Type object4(nullptr); // Test nullptr (see CPPPSM-74)
  #endif

    RTI_TEST_ASSERT(object1.delegate().use_count() == 2);
    RTI_TEST_ASSERT(object2.delegate().use_count() == 1);
    RTI_TEST_ASSERT(object3.delegate().use_count() == 2);
    RTI_TEST_ASSERT(object4.delegate().get() == NULL);

    object4 = object3 = object1;

    RTI_TEST_ASSERT(object1.delegate().use_count() == 3);
    RTI_TEST_ASSERT(object2.delegate().use_count() == 1);
    RTI_TEST_ASSERT(object3.delegate().use_count() == 3);
    RTI_TEST_ASSERT(object4.delegate().use_count() == 3);

    RTI_TEST_ASSERT(object1 == object3);
    RTI_TEST_ASSERT(native_getter(object1) == native_getter(object3));
    RTI_TEST_ASSERT(native_getter(object1) != native_getter(object2));

    object1 = object2;
    RTI_TEST_ASSERT(native_getter(object1) != native_getter(object3));
    RTI_TEST_ASSERT(native_getter(object1) == native_getter(object2));

    // Assignment to reference of base class
    BaseType entity = object3;
    RTI_TEST_ASSERT(
        native_getter(entity) == native_getter(object3));

    object3 = dds::core::null;
    RTI_TEST_ASSERT(object3.delegate().get() == NULL);
    RTI_TEST_ASSERT(object3 == dds::core::null);

    // Test polymorphic downcast from Entity to T
    object3 = dds::core::polymorphic_cast<Type>(entity);
    RTI_TEST_ASSERT(object3.delegate().get() != NULL);
    RTI_TEST_ASSERT(
        native_getter(entity) == native_getter(object3));
    RTI_TEST_ASSERT(
        dynamic_cast<typename Type::DELEGATE_T*>(entity.delegate().get()));
    object3 = dds::core::null;

    entity = dds::core::null;
    RTI_TEST_ASSERT(entity.delegate().get() == NULL);
    RTI_TEST_ASSERT(entity == dds::core::null);
  #ifndef RTI_NO_CXX11_NULLPTR
    RTI_TEST_ASSERT(entity == nullptr);
  #endif

    test_entity_in_containers<BaseType>(object1, object4);
    test_entity_swap(object1, object4);
}

struct NativeEntityGetter {
    const DDS_Entity * operator() (const dds::core::Entity& entity)
    {
        return entity->native_entity();
    }
};

/*
 * Test the reference semantics of entity classes (T is a class derived
 * from dds::core::Entity). Test that assignment to the base class is possible.
 *
 * @pre There are no other references to object1 and object2
 */
template <class EntityType>
void test_entity_reference_assignment(EntityType& object1, EntityType& object2)
{
    test_reference_assignment<EntityType, dds::core::Entity, NativeEntityGetter>(
        object1, object2);
}

template <typename Type,
          typename NativeType = typename rti::core::native_type_traits<Type>::native_type,
          typename TypeAdapter = typename rti::core::native_type_traits<Type>::adapter_type >
struct test_native_type_memory_mapping {
    // This is a struct instead of a standalone function because until C++11,
    // a template function can't have default template arguments
    static void run()
    {
        RTI_TEST_ASSERT(sizeof(Type) == sizeof(NativeType));

        Type cpp_variable;
        NativeType c_variable;

        TypeAdapter::initialize(c_variable);
        RTI_TEST_ASSERT(memcmp(&c_variable, &cpp_variable, sizeof(Type)) == 0);
    }
};

#ifndef RTI_NO_CXX11_SMART_PTR
template<typename Listener, typename Entity>
void test_entity_listener(Entity entity)
{
    RTI_TEST_ASSERT(entity.get_listener() == nullptr);

    // Set listener
    auto listener = std::make_shared<Listener>();
    entity.set_listener(listener);
    RTI_TEST_ASSERT(entity.get_listener() == listener);
    // The deprecated function listener() still works
    RTI_TEST_ASSERT(entity.listener() == listener.get());

    auto listener2 = std::make_shared<Listener>();
    entity.set_listener(listener2);
    RTI_TEST_ASSERT(entity.get_listener() == listener2);
    RTI_TEST_ASSERT(entity.listener() == listener2.get());

    // Calling the deprecated listener() setter when a shared_ptr listener has
    // been set with set_listener() is illegal
    RTI_TEST_ASSERT_EXCEPTION_NO_WARN(
            entity.listener(NULL, dds::core::status::StatusMask::none()),
            dds::core::IllegalOperationError);

    // Set null listener
    entity.set_listener(nullptr);
    RTI_TEST_ASSERT(entity.get_listener() == nullptr);
    RTI_TEST_ASSERT(entity.listener() == nullptr);
}
#endif

} } // namespace rti::test

#ifdef TARGET
    #define RTI_TEST_SKIP_TARGET(target__)                                \
        if (std::string(TARGET) == target__) {                            \
            RTITestLog_freeForm(                                          \
                    RTI_LOG_BIT_EXCEPTION,                                \
                    "SKIPPED - test not supported on this platform\n");   \
            return;                                                       \
        }
#else
    #define RTI_TEST_SKIP_TARGET(target)
#endif

#endif // RTI_TEST_UTILITIES_HPP_

/* End $Id$ */
