/*
 Copyright (c) 2021, Real-Time Innovations, Inc. All rights reserved.

 No duplications, whole or partial, manual or electronic, may be made
 without express written permission.  Any such copies, or
 revisions thereof, must display this notice unaltered.
 This code contains trade secrets of Real-Time Innovations, Inc.
*/

#ifndef RTI_DDS_TOPIC_CDR_GENERICINTERPRETERTYPEPLUGIN_HPP_
#define RTI_DDS_TOPIC_CDR_GENERICINTERPRETERTYPEPLUGIN_HPP_

#include <dds/core/macros.hpp>

#include <string.h>
#include <memory>

#include "ndds/ndds_c.h"
#include "osapi/osapi_type.h"
#include "osapi/osapi_heap.h"
#include "osapi/osapi_utility.h"
#include "cdr/cdr_type.h"
#include "cdr/cdr_typeObject.h"
#include "cdr/cdr_encapsulation.h"
#include "cdr/cdr_stream.h"
#include "xcdr/xcdr_interpreter.h"
#include "xcdr/xcdr_stream.h"
#include "cdr/cdr_log.h"
#include "pres/pres_typePlugin.h"
#include "dds_c/dds_c_typecode_impl.h"

#include <dds/domain/DomainParticipant.hpp>
#include <dds/topic/Topic.hpp>
#include "dds/topic/TopicTraits.hpp"

namespace rti { namespace topic { namespace cdr {

// A type plugin that uses the interpreter to initialize and serialize data
// samples for a type defined by a DynamicType with a default C data layout.
//
// GenericTypePlugin has two uses:
//
// 1) It creates a PRESTypePlugin with all the callbacks required to process
//    data samples in the writer and reader queues. These are static methods.
//
// 2) It provides non-static methods that can be used by a language binding to
//    manipulate data samples and perform actions such as allocating and
//    initializing an optional member or resizing a sequence.
//
// GenericTypePlugin is currently used by the Python API. It uses it to
// serialize and deserialize C-like data samples (1) and convert them from
// C to Python and the other way around (2).
//
// SampleWrapper is simply the type that holds the data. For
// a given SampleWrapper instance sw, the expression get_sample(sw) must compile
// and return a void* pointing to the data sample. The default SampleWrapper
// is simply a pointer to the sample itself.
//
template<typename SampleWrapper>
class GenericTypePlugin {
public:
    GenericTypePlugin(dds::core::xtypes::DynamicType &type);

    // This type can't be copied
    GenericTypePlugin(const GenericTypePlugin &) = delete;
    GenericTypePlugin(GenericTypePlugin &&) = default;
    GenericTypePlugin &operator=(const GenericTypePlugin &) = delete;
    GenericTypePlugin &operator=(GenericTypePlugin &&) = default;

    ~GenericTypePlugin();

    const std::string& type_name() const
    {
        return type_name_;
    }

    void register_type(
            dds::domain::DomainParticipant &participant,
            const char *name);

    void register_type(dds::domain::DomainParticipant &participant);

    unsigned int serialized_sample_size(
            const SampleWrapper &sample,
            DDS_DataRepresentationId_t representation =
                    DDS_AUTO_DATA_REPRESENTATION);

    unsigned int serialized_sample_max_size(
            DDS_DataRepresentationId_t representation =
                    DDS_AUTO_DATA_REPRESENTATION);

    unsigned int serialize_to_cdr_buffer(
            std::vector<char> &buffer,
            const SampleWrapper &sample,
            DDS_DataRepresentationId_t representation =
                    DDS_AUTO_DATA_REPRESENTATION,
            bool keys_only = false);

    unsigned int serialize_to_cdr_buffer(
            char *buffer,
            unsigned int buffer_size,
            const SampleWrapper &sample,
            DDS_DataRepresentationId_t representation =
                    DDS_AUTO_DATA_REPRESENTATION,
            bool keys_only = false);

    void deserialize_from_cdr_buffer(
            SampleWrapper &sample,
            const char *buffer,
            unsigned int length,
            bool keys_only = false,
            bool sample_reused = false);

    void deserialize_from_cdr_buffer(
            SampleWrapper &sample,
            const std::vector<char> &buffer,
            bool keys_only = false,
            bool sample_reused = false);

    // Initializes a data sample using the interpreter programs for this type
    void initialize_sample(SampleWrapper &sample);

    // Initializes a particular member -- used for optional members
    void initialize_member(SampleWrapper &sample, uint32_t instruction_index);

    // Resizes a member -- used for optional sequences
    void resize_member(
            SampleWrapper &sample,
            uint32_t instruction_index,
            uint32_t new_size);
    void finalize_sample(SampleWrapper &sample);

    // Finalizes and nullifies all optional members at any level of nestedness.
    void finalize_optional_members(SampleWrapper &sample);

    const RTIXCdrProgram *get_initialization_program() const
    {
        return RTIXCdrInterpreterPrograms_getInitializeSampleProgram(
                xcdr_programs_);
    }

    const RTIXCdrProgram *get_finalization_program() const
    {
        return RTIXCdrInterpreterPrograms_getAllocatedMembersSampleProgram(
                xcdr_programs_);
    }

    // Returns the instruction index (required by resize_member and
    // initialize_member) from the index of the member in the type. For unions,
    // 0 is the discriminator and 1 is the index of the first case.
    uint32_t get_instruction_index_from_member_index(
            uint32_t member_index) const;

private:
    using KeyHolder = SampleWrapper;

    RTIXCdrInterpreterPrograms *create_programs();

    RTIXCdrInterpreterPrograms *get_programs()
    {
        return xcdr_programs_;
    }

    const RTIXCdrInterpreterPrograms *get_programs() const
    {
        return xcdr_programs_;
    }

    void setup_endpoint_data_for_cdr_buffer(
            PRESTypePluginDefaultEndpointData &epd,
            PRESTypePluginDefaultParticipantData &pd);

    void check_valid_program_mask(
            RTIXCdrProgramMask requested_mask,
            const char *program_name);

    static PRESTypePluginParticipantData on_participant_attached(
            void *registration_data,
            const struct PRESTypePluginParticipantInfo *participant_info,
            RTIBool top_level_registration,
            void *container_plugin_context,
            RTICdrTypeCode *typeCode);

    static void on_participant_detached(
            PRESTypePluginParticipantData participant_data);

    static PRESTypePluginEndpointData on_endpoint_attached(
            PRESTypePluginParticipantData participant_data,
            const struct PRESTypePluginEndpointInfo *endpoint_info,
            RTIBool top_level_registration,
            void *container_plugin_context);

    static void on_endpoint_detached(PRESTypePluginEndpointData endpoint_data);

    static void return_sample(
            PRESTypePluginEndpointData endpoint_data,
            SampleWrapper *sample,
            void *handle);

    static RTIBool copy_sample(
            PRESTypePluginEndpointData endpoint_data,
            SampleWrapper *out,
            const SampleWrapper *in);

    static RTIBool serialize(
            PRESTypePluginEndpointData endpoint_data,
            const SampleWrapper *sample,
            struct RTICdrStream *stream,
            RTIBool serialize_encapsulation,
            RTIEncapsulationId encapsulation_id,
            RTIBool serialize_sample,
            void *endpoint_plugin_qos);

    static RTIBool deserialize_sample(
            PRESTypePluginEndpointData endpoint_data,
            SampleWrapper *sample,
            struct RTICdrStream *stream,
            RTIBool deserialize_encapsulation,
            RTIBool deserialize_sample,
            void *endpoint_plugin_qos);

    static RTIBool deserialize(
            PRESTypePluginEndpointData endpoint_data,
            SampleWrapper **sample,
            RTIBool *drop_sample,
            struct RTICdrStream *stream,
            RTIBool deserialize_encapsulation,
            RTIBool deserialize_sample,
            void *endpoint_plugin_qos);

    static unsigned int get_serialized_sample_max_size(
            PRESTypePluginEndpointData endpoint_data,
            RTIBool include_encapsulation,
            RTIEncapsulationId encapsulation_id,
            unsigned int current_alignment);

    static unsigned int get_serialized_sample_size(
            PRESTypePluginEndpointData endpoint_data,
            RTIBool include_encapsulation,
            RTIEncapsulationId encapsulation_id,
            unsigned int current_alignment,
            const SampleWrapper *sample);

    static PRESTypePluginKeyKind get_key_kind_nokey();
    static PRESTypePluginKeyKind get_key_kind_key();

    static unsigned int get_serialized_key_max_size(
            PRESTypePluginEndpointData endpoint_data,
            RTIBool include_encapsulation,
            RTIEncapsulationId encapsulation_id,
            unsigned int current_alignment);

    static unsigned int get_serialized_key_max_size_for_keyhash(
            PRESTypePluginEndpointData endpoint_data,
            RTIEncapsulationId encapsulation_id,
            unsigned int current_alignment);

    static RTIBool serialize_key(
            PRESTypePluginEndpointData endpoint_data,
            const SampleWrapper *sample,
            struct RTICdrStream *stream,
            RTIBool serialize_encapsulation,
            RTIEncapsulationId encapsulation_id,
            RTIBool serialize_key,
            void *endpoint_plugin_qos);

    static RTIBool deserialize_key_sample(
            PRESTypePluginEndpointData endpoint_data,
            SampleWrapper *sample,
            struct RTICdrStream *stream,
            RTIBool deserialize_encapsulation,
            RTIBool deserialize_key,
            void *endpoint_plugin_qos);

    static RTIBool deserialize_key(
            PRESTypePluginEndpointData endpoint_data,
            SampleWrapper **sample,
            RTIBool *drop_sample,
            struct RTICdrStream *stream,
            RTIBool deserialize_encapsulation,
            RTIBool deserialize_key,
            void *endpoint_plugin_qos);

    static RTIBool serialized_sample_to_key(
            PRESTypePluginEndpointData endpoint_data,
            SampleWrapper *sample,
            struct RTICdrStream *stream,
            RTIBool deserialize_encapsulation,
            RTIBool deserialize_key,
            void *endpoint_plugin_qos);

    static RTIBool instance_to_keyhash(
            PRESTypePluginEndpointData endpoint_data,
            DDS_KeyHash_t *keyhash,
            const SampleWrapper *instance,
            RTIEncapsulationId encapsulationId);

    static RTIBool serialized_sample_to_keyhash(
            PRESTypePluginEndpointData endpoint_data,
            struct RTICdrStream *stream,
            DDS_KeyHash_t *keyhash,
            RTIBool deserialize_encapsulation,
            void *endpoint_plugin_qos);

    static void finalize_optional_members(
            PRESTypePluginEndpointData endpoint_data,
            void *sample,
            RTIBool finalize);

    // --- Data creation functions:

    // This function is called by the PRES reader queue to create new samples
    static RTIBool create_sample(void *buffer, void *param) noexcept;

    // This function is called by the XCDR interpreter to set a sample back
    // to its default values before deserialization.
    static RTIXCdrLong initialize_xcdr_sample(
            void *sample,
            const struct RTIXCdrTypeCode *typeCode,
            RTIXCdrUnionInitializeInfo *,
            void *programData,
            void *param);
    static RTIXCdrLong initialize_xcdr_union_sample(
            void *sample,
            const struct RTIXCdrTypeCode *typeCode,
            RTIXCdrUnionInitializeInfo *,
            void *programData,
            void *param);

    static RTIXCdrLong finalize_xcdr_sample(
            void *sample,
            const RTIXCdrTypeCode *typeCode,
            void *programData,
            void *param);

    // This function is called by the PRES reader queue to delete a sample from
    // the queue.
    static void finalize_sample_callback(void *buffer, void *param);

    static RTIBool create_key(void *buffer, void *param)
    {
        return create_sample(buffer, param);
    }

    static void finalize_key(void *buffer, void *param)
    {
        finalize_sample_callback(buffer, param);
    }

private:
    std::string type_name_;
    PRESTypePlugin plugin_;
    RTIXCdrTypePlugin xcdr_plugin_;
    RTIXCdrInterpreterPrograms *xcdr_programs_;
    RTIXCdrProgramMask xcdr_program_mask_ = RTI_XCDR_PROGRAM_MASK_NONE;
};

template<typename SampleWrapper>
dds::topic::Topic<SampleWrapper> create_topic(
        dds::domain::DomainParticipant participant,
        const std::string &topic_name,
        const std::string &type_name,
        GenericTypePlugin<SampleWrapper> &type_plugin)
{
    // Register the type plugin. Unlike a regular Topic<Foo> where a unique
    // topic_type_support<Foo> can be accessed at compile-time, here we need the
    // input type_plugin for each type.
    type_plugin.register_type(participant, type_name.c_str());

    // Bypass the public dds::topic::Topic constructors, which require a regular
    // C++ Foo type, or DynamicData.
    return dds::topic::Topic<SampleWrapper>(
            dds::core::construct_from_native_tag_t(),
            new rti::topic::TopicImpl<SampleWrapper>(
                    rti::topic::detail::no_register_tag_t(),
                    participant,
                    topic_name,
                    type_name,
                    NULL,
                    nullptr,
                    dds::core::status::StatusMask::none()));
}


template<typename SampleWrapper>
dds::topic::Topic<SampleWrapper> create_topic(
        dds::domain::DomainParticipant participant,
        const std::string &topic_name,
        const std::string &type_name,
        GenericTypePlugin<SampleWrapper> &type_plugin,
        const dds::topic::qos::TopicQos &qos,
        std::shared_ptr<dds::topic::TopicListener<SampleWrapper>> listener =
                nullptr,
        const dds::core::status::StatusMask &mask =
                dds::core::status::StatusMask::all())
{
    // Register the type plugin. Unlike a regular Topic<Foo> where a unique
    // topic_type_support<Foo> can be accessed at compile-time, here we need the
    // input type_plugin for each type.
    type_plugin.register_type(participant, type_name.c_str());

    // Bypass the public dds::topic::Topic constructors, which require a regular
    // C++ Foo type, or DynamicData.
    return dds::topic::Topic<SampleWrapper>(
            dds::core::construct_from_native_tag_t(),
            new rti::topic::TopicImpl<SampleWrapper>(
                    rti::topic::detail::no_register_tag_t(),
                    participant,
                    topic_name,
                    type_name,
                    &qos,
                    listener,
                    mask));
}


// The simplest SampleWrapper; it works for IDL C types.
// This empty struct is only used as the type for topics that use the generic
// type plugin. We need a distinct type instead of "void"
//
// - CSampleWrapper& is equivalent to "void"
// - CSampleWrapper* is equivalent to "void*"
//
// The plugin simply casts objects of this type to void*
//
struct CSampleWrapper {
    // Objects of this type can never be created directly or assigned; they are
    // always cast from a void*.
    CSampleWrapper() = delete;
    CSampleWrapper(const CSampleWrapper&) = delete;
    CSampleWrapper(CSampleWrapper&&) = delete;
    CSampleWrapper& operator=(const CSampleWrapper&) = delete;
    CSampleWrapper& operator=(CSampleWrapper&&) = delete;

    void *sample()
    {
        // This object and the sample are the same thing
        return this;
    }

    const void *sample() const
    {
        return this;
    }
};

inline void *get_sample(CSampleWrapper *wrapper)
{
    return wrapper;
}

inline const void *get_sample(const CSampleWrapper *wrapper)
{
    return wrapper;
}

// This type plugin works for any C type given a properly initialized
// DynamicType with sample access info.
using CTypePlugin = GenericTypePlugin<CSampleWrapper>;

// Enables XML Application Creation for CTypePlugin-based types. This function
// provides the CTypePlugin so that it can be used by participants to
// register CTypePlugin-based types with the given registered_type_name.
//
// FactoryPluginPolicy is a type modeled after rti::domain::FactoryPluginSupport
// that configures DDS_FactoryPluginSupport with the right callbacks. For a C++
// type FactoryPluginPolicy can be set to rti::domain::FactoryPluginSupport.
template <typename FactoryPluginPolicy>
void register_c_type_plugin(
        CTypePlugin *plugin,
        const std::string& registered_type_name)
{
    DDS_ReturnCode_t retcode = DDS_RETCODE_ERROR;

    if (plugin == nullptr) {
        throw dds::core::InvalidArgumentError(
                "plugin cannot be null");
    }

    const std::lock_guard<std::mutex> lock_guard(
            rti::domain::FactoryPluginSupport::mutex());

    rti::domain::detail::set_factory_plugin_support<FactoryPluginPolicy>();

    // This holder is immediately retrieved by
    // DDS_DomainParticipantFactory_register_type_supportI before the function
    // ends.
    rti::core::UserProxTypeSupportRegisterTypeHolder holder(
            [plugin](
                    dds::domain::DomainParticipant& participant,
                    const std::string& type_name) {
                plugin->register_type(participant, type_name.c_str());
            });

    retcode = DDS_DomainParticipantFactory_register_type_supportI(
            DDS_DomainParticipantFactory_get_instance(),
            reinterpret_cast<void *>(&holder),
            registered_type_name.c_str());
    rti::core::check_return_code(retcode, "register_type_support");
}

}}}  // namespace rti::topic::cdr

namespace dds { namespace topic {

// Enable CSampleWrapper as a topic-type, a valid template argument for Topic<T>.
template<>
struct is_topic_type<rti::topic::cdr::CSampleWrapper> : std::true_type {
};

}}  // namespace dds::topic

namespace rti { namespace topic {

// Disable operations that require copying the data in C++, such as "take_next";
// CSampleWrapper doesn't have a copy constructor that deeply copies the data.
template<>
struct topic_type_disabled_copy<rti::topic::cdr::CSampleWrapper>
        : std::true_type {
};

}} // namespace rti::topic

namespace dds { namespace topic {
template <>
struct topic_type_name<rti::topic::cdr::CSampleWrapper> {
    static std::string value()
    {
        static const std::string name = "CSampleWrapper";
        return name;
    }
};
}}  // namespace dds::topic

#endif  // RTI_DDS_TOPIC_CDR_GENERICINTERPRETERTYPEPLUGIN_HPP_
