/* $Id$

 (c) Copyright, Real-Time Innovations, 2022
 All rights reserved.
 No duplications, whole or partial, manual or electronic, may be made
 without express written permission.  Any such copies, or
 revisions thereof, must display this notice unaltered.
 This code contains trade secrets of Real-Time Innovations, Inc.

=========================================================================*/

#ifndef monitoring_monitoringForwarder_h
#define monitoring_monitoringForwarder_h

#ifndef monitoring_dll_h
  #include "monitoring/monitoring_dll.h"
#endif
#ifndef ndds_c_h
  #include "ndds/ndds_c.h"
#endif
#ifndef dds_c_monitoring_h
  #include "dds_c/dds_c_monitoring.h"
#endif
#ifndef reda_inlineMemory_h
  #include "reda/reda_inlineMemory.h"
#endif

#include "monitoring/monitoring_dataModel.h"

#include "dds_logging.h"

#ifdef __cplusplus
    extern "C" {
#endif

/*i
 * @brief The Monitoring Application Resource type contains information about
 * the underlying monitoring resource plus logging configuration settings (which
 * are stored here as a proxy for remote controllability).
 */
struct RTI_MonitoringForwarderApplicationResource {
    /* The logging configuration of the application */
    DDS_Monitoring_LoggingConfig loggingConfig;
    /* The Fully qualified name to use when matching a resource expression */
    char *fullyQualifiedName;
};

typedef struct RTI_MonitoringForwarderApplicationResource *
        RTI_MonitoringForwarderApplicationResourcePtr;

/*i
 * @brief The Monitoring Forwarder needs a way to return a collection of
 * Application Resources when looking for Application names that match a certain
 * selector expression. We declare and implement the DDS Sequence pattern here
 * for pointers to Application Resources.
 *
 * @see RTI_MonitoringForwarder_find_matching_applications
 */
DDS_SEQUENCE(
        RTI_MonitoringForwarderApplicationResourcePtrSeq,
        RTI_MonitoringForwarderApplicationResourcePtr);

/*i
 * @brief Monitoring Forwarder Entities.
 * It contains the pointers to the readers required in order to receive data 
 * from the Monitoring Library. These readers provide consistent information
 * to the user of the Monitoring Forwarder. For the reliable topics, the
 * forwarder will take care of processing the samples and requesting for
 * snapshots, if needed, before providing the samples to these readers.
 */
struct RTI_MonitoringForwarderEntities {
    DDS_DataReader *event_reader;
    DDS_DataReader *periodic_reader;
    DDS_DataReader *logging_reader;
};

/*i
 * @brief Monitoring Forwarder Class.
 * A Monitoring forwarder is used to read from the monitoring topics and
 * to send commands that control the process of monitoring. Therefore, it is
 * composed of a set of readers and an RPC instance.
 */
struct RTI_MonitoringForwarder {
    DDS_DomainParticipantFactory *factory;
    struct RTI_MonitoringForwarderEntities *entities;
};

/*i @brief The property structure to configure the Monitoring Forwarder */
struct RTI_MonitoringForwarderProperty_t {
    /*i 
     * The domain ID used by the forwarder, it only applies if the
     * DomainParticipant is created by the forwarder.
     */
    DDS_DomainId_t domain_id;
    /*i 
     * The DomainParticipant QoS used by the forwarder's DomainParticipant,
     * it only applies if the DomainParticipant is created by the forwarder.
     */
    struct DDS_DomainParticipantQos *participant_qos;
    /*i 
     * The subscriber Qos used by every builtin subscriber of the forwarder.
     * If NULL, the forwarder will use the Monitoring2 generic QoS 
     * profile (\idref_PROFILE_GENERIC_MONITORING2).
     */
    struct DDS_SubscriberQos *subscriber_qos;
    /*i 
     * The DataReader Qos used by the reader for the event topic.
     * If NULL, the forwarder will use the Monitoring2 generic QoS 
     * profile (\idref_PROFILE_GENERIC_MONITORING2).
     */
    struct DDS_DataReaderQos *event_reader_qos;
    /*i 
     * The DataReader Qos used by the reader for the periodic topic.
     * If NULL, the forwarder will use the Monitoring2 generic QoS 
     * profile (\idref_PROFILE_GENERIC_MONITORING2).
     */
    struct DDS_DataReaderQos *periodic_reader_qos;
    /*i 
     * The DataReader Qos used by the reader for the logging topic.
     * If NULL, the forwarder will use the Monitoring2 generic QoS 
     * (\idref_PROFILE_GENERIC_MONITORING2).
     */
    struct DDS_DataReaderQos *logging_reader_qos;
    /*i Timeout duration for commands */
    struct DDS_Duration_t command_manager_timeout;
    /*i
     * Whether to create or not the logging DataReader.
     * This should be set to false, for example, if you are using security and
     * your permissions file doesn't allow subscribing to logging.
     */
    DDS_Boolean create_logging_data_reader;
};

/*i Monitoring Forwarder property initializer
 */
#define RTI_MonitoringForwarderProperty_t_INITIALIZER { \
    2, /* domain_id */ \
    NULL, /* participant_qos */ \
    NULL, /* subscriber_qos */ \
    NULL, /* event_reader_qos */ \
    NULL, /* periodic_reader_qos */ \
    NULL, /* logging_reader_qos */ \
    {10, 0}, /* command_manager_timeout */ \
    DDS_BOOLEAN_TRUE /* create_logging_data_reader */ \
}

/*i
 * @brief Finalizes a RTI_MonitoringForwarderProperty_t instance. 
 * @param self \b In. The RTI_MonitoringForwarderProperty_t that is going to
 * be finalized.
 */
extern RTI_MonitoringDllExport
void RTI_MonitoringForwarderProperty_t_finalize(
        struct RTI_MonitoringForwarderProperty_t *self);

/*i
 * @brief Create an instance of RTI_MonitoringForwarder.
 * @param factory \b In. The DDS_DomainParticipantFactory used by the
 * RTI_MonitoringForwarder instance to create DDS entities.
 * @param property \b In. The property for the Monitoring Library Forwarder.
 * @return A Monitoring Forwarder instance or NULL if there is an error.
 */
extern RTI_MonitoringDllExport
struct RTI_MonitoringForwarder * RTI_MonitoringForwarder_new(
        DDS_DomainParticipantFactory *factory,
        const struct RTI_MonitoringForwarderProperty_t *property);

/*i
 * @brief Create an instance of RTI_MonitoringForwarder using the
 * DomainParticipant provided instead of creating a new one.
 * @param participant \b In. The DDS_DomainParticipant used by the
 * RTI_MonitoringForwarder, the other entities required by the forwarder will
 * be created from this DomainParticipant.
 * @param property \b In. The property for the Monitoring Library Forwarder.
 * @return A Monitoring Forwarder instance or NULL if there is an error.
 */
extern RTI_MonitoringDllExport
struct RTI_MonitoringForwarder * RTI_MonitoringForwarder_new_w_participant(
        DDS_DomainParticipant *participant,
        const struct RTI_MonitoringForwarderProperty_t *property);

/*i
 * @brief Delete the Monitoring Library Forwarder instance.
 * @param self \b In. The Monitoring Library Forwarder instance to remove.
 */
extern RTI_MonitoringDllExport
void RTI_MonitoringForwarder_delete(
        struct RTI_MonitoringForwarder *self);

/*i
 * @brief Converts a MonitoringLogging message to its string representation.
 * @details This method will do the following to message->message:
 * - Add the timestamp in UTC format if it is not already present.
 * - Add the Facility.
 * - Replace the Connext LogLevel with the SyslogLevel.
 * - Expand these GUIDs that don't have the prefix.
 * - Remove every occurrence of "0x" in the text (even if it doesn't belong
 *   to a GUID).
 * - Replace the commas and colons in the GUIDs with dots.
 * - Replace newline characters (\n) with spaces, except the trailing newline.
 * - If the input message doesn't have a trailing newline, add it.
 * @param message \b In. The input Monitoring LoggingMessage struct.
 * @param buffer \b InOut. The buffer in which the string representation of
 * message will be stored. Its existing content will be overwritten.
 * @param size \b buffer's size.
 * @attention Upon failure, buffer will contain a copy of message->message.
 */
extern RTI_MonitoringDllExport
void RTI_MonitoringForwarder_log_message_to_string(
        const DDS_Monitoring_LoggingMessage *message,
        char *buffer,
        size_t size);


/*i
 * @brief Sends a command to set the forwarding level of one or many logging
 * groups
 * @param self \b InOut. The forwarder that receives the command
 * @param exception \b Out. A DDS_Monitoring_MonitoringCommandError pointer
 * where to store the output of the command. This structure only stores if there
 * was an error or not and a string related to the error. The actual response of
 * the command, if any, will be received through the Monitoring Topics. 
 * @param logging_groups \b In The sequence of logging groups modified by this
 * command.
 */
extern RTI_MonitoringDllExport
void RTI_MonitoringForwarder_set_logging_forwarding_level(
        struct RTI_MonitoringForwarder *self,
        DDS_Monitoring_MonitoringCommandError *exception,
        const DDS_Monitoring_LoggingGroupSequence *logging_groups);

/*i
 * @brief Sends a command to set the collection level of one or many logging
 * groups
 * @param self \b InOut. The forwarder that receives the command
 * @param exception \b Out. A DDS_Monitoring_MonitoringCommandError pointer
 * where to store the output of the command. This structure only stores if there
 * was an error or not and a string related to the error. The actual response of
 * the command, if any, will be received through the Monitoring Topics. 
 * @param logging_groups \b In The sequence of logging groups modified by this
 * command.
 */
extern RTI_MonitoringDllExport
void RTI_MonitoringForwarder_set_logging_collection_level(
        struct RTI_MonitoringForwarder *self,
        DDS_Monitoring_MonitoringCommandError *exception,
        const DDS_Monitoring_LoggingGroupSequence *logging_groups);

/*i
 * @brief Sends a command to get the subscription state of an entity
 * @param self \b InOut. The forwarder that receives the command
 * @param metrics \b Out A DDS_Monitoring_MetricSelectorSequence pointer
 * to store the metric subscription state output for the command.
 * @param exception \b Out. A DDS_Monitoring_MonitoringCommandError pointer
 * where to store the output of the command. This structure only stores if there
 * was an error or not and a string related to the error. 
 * @param resources \b In The sequence of resources to get subscription
 * state for.
 */
extern RTI_MonitoringDllExport
void RTI_MonitoringForwarder_get_metric_subscription_state(
        struct RTI_MonitoringForwarder *self,
        DDS_Monitoring_MetricSelectorSequence *metrics,
        DDS_Monitoring_MonitoringCommandError *exception,
        const DDS_Monitoring_ResourcePathExpressionSequence *resources);

/*i
 * @brief Sends a command to set the subscription state for one or more
 * entities.
 * @param self \b InOut. The forwarder that receives the command
 * @param exception \b In. A DDS_Monitoring_MonitoringCommandError pointer
 * where to store the output of the command. This structure only stores if there
 * was an error or not and a string related to the error. 
 * @param metrics \b In The sequence of metrics to subscribe to. Metrics not in
 * the sequence for an entity will be unsubscribed.
 */
extern RTI_MonitoringDllExport
void RTI_MonitoringForwarder_set_metric_subscription_state(
        struct RTI_MonitoringForwarder *self,
        DDS_Monitoring_MonitoringCommandError *exception,
        const DDS_Monitoring_MetricSelectorSequence *metrics);

/*i
 * @brief Sends a command to update the subscription state for one or more
 * entities.
 * @param self \b InOut. The forwarder that receives the command
 * @param exception \b In. A DDS_Monitoring_MonitoringCommandError pointer
 * where to store the output of the command. This structure only stores if there
 * was an error or not and a string related to the error.
 * @param metrics \b In The sequence of resources and metrics for which the 
 * metrics subscription will be updated. The metrics specified in the
 * subscribe_metric_selector field will be subscribed to while metrics in the
 * unsubscribe_metric_selector field will be unsubscribed. Subscription state
 * for any metrics not specified in one of the lists will not be changed.
 */
extern RTI_MonitoringDllExport
void RTI_MonitoringForwarder_update_metric_subscription_state(
        struct RTI_MonitoringForwarder *self,
        DDS_Monitoring_MonitoringCommandError *exception,
        const DDS_Monitoring_MetricSelectorSequence *metrics);

/*i
 * @brief Locks the application registry of the forwarder in order to perform
 * search operations on it.
 * @param self \b InOut. The forwarder object that will be locked.
 * @return DDS_BOOLEAN_TRUE if the lock happened successfully, DDS_BOOLEAN_FALSE
 * otherwise.
 */
extern RTI_MonitoringDllExport
DDS_Boolean RTI_MonitoringForwarder_lock_application_registry(
        struct RTI_MonitoringForwarder *self);

/*i
 * @brief Unlocks the application registry of the forwarder after performing
 * search operations on it.
 * @param self \b InOut. The forwarder object that will be unlocked.
 * @return DDS_BOOLEAN_TRUE if the unlock happened successfully,
 * DDS_BOOLEAN_FALSE otherwise.
 */
extern RTI_MonitoringDllExport
DDS_Boolean RTI_MonitoringForwarder_unlock_application_registry(
        struct RTI_MonitoringForwarder *self);

/*i
 * @brief Retrieves a sequence of applications that match a given expression.
 * This call and the usage of its output parameter needs to be enclosed between
 * calls to RTI_MonitoringForwarder_lock_application_registry and 
 * RTI_MonitoringForwarder_unlock_application_registry for it to be safe.
 * @param self \b In. The Monitoring Forwarder object.
 * @param app_seq \b Out. The sequence of applications to populate with the
 * monitored applications matching the expression.
 * @param resource_name_expr \b In. The resource expression the applications have
 * to match.
 * @return DDS_BOOLEAN_TRUE if the list of applications was obtained
 * successfully, DDS_BOOLEAN_FALSE otherwise. The sequence of applications can
 * be empty if no application matched the expression. That is not considered an
 * error and the function will return DDS_BOOLEAN_TRUE in that case.
 */
extern RTI_MonitoringDllExport
DDS_Boolean RTI_MonitoringForwarder_find_matching_applications(
        const struct RTI_MonitoringForwarder *self,
        struct RTI_MonitoringForwarderApplicationResourcePtrSeq *app_seq,
        const char *resource_name_expr);

#ifdef __cplusplus
    }	/* extern "C" */
#endif

#endif /* monitoring_monitoringForwarder_h */
/* end of $Id$ */