/* $Id$

 (c) Copyright, Real-Time Innovations, 2022
 All rights reserved.
 No duplications, whole or partial, manual or electronic, may be made
 without express written permission.  Any such copies, or
 revisions thereof, must display this notice unaltered.
 This code contains trade secrets of Real-Time Innovations, Inc.

=========================================================================*/

#ifndef monitoring_monitoringWriterEntities_h
#define monitoring_monitoringWriterEntities_h

#ifndef monitoring_dll_h
  #include "monitoring/monitoring_dll.h"
#endif
#ifndef ndds_c_h
  #include "ndds/ndds_c.h"
#endif

#ifdef __cplusplus
    extern "C" {
#endif

/*i
 * @brief Monitoring Writer Entities.
 * It contains the pointers to the built-in writers required in order to send 
 * monitoring telemetry data (event metrics, periodic metrics, and logging).
 */
struct RTI_MonitoringWriterEntities {
    DDS_DataWriter *event_writer;
    DDS_DataWriter *periodic_writer;
    DDS_DataWriter *logging_writer;
};

/*i
 * @brief The property structure used to configure the monitoring writer
 * entities
 */
struct RTI_MonitoringWriterEntitiesProperty_t {
    /*i
     * @brief The publisher Qos used to create the publishers for the event,
     * periodic and logging topics.
     * 
     * If NULL, the forwarder will use the Monitoring2 generic QoS 
     * profile (\idref_PROFILE_GENERIC_MONITORING2).
     */
    struct DDS_PublisherQos *publisher_qos;
    /*i
     * @brief The DataWriter Qos used to create the writer for the event topic.
     *
     * If NULL, the forwarder will use the Monitoring2 generic QoS 
     * profile (\idref_PROFILE_GENERIC_MONITORING2).
     */
    struct DDS_DataWriterQos *event_writer_qos;
    /*i
     * @brief The DataWriter Qos used to create the writer for the periodic
     * topic.
     * 
     * If NULL, the forwarder will use the Monitoring2 generic QoS 
     * profile (\idref_PROFILE_GENERIC_MONITORING2).
     */
    struct DDS_DataWriterQos *periodic_writer_qos;
    /*i
     * @brief The DataWriter Qos used to create the writer for the logging
     * topic.
     * 
     * If NULL, the forwarder will use the Monitoring2 generic QoS 
     * profile (\idref_PROFILE_GENERIC_MONITORING2).
     */
    struct DDS_DataWriterQos *logging_writer_qos;
};

/*i
 * Monitoring Initializer for the monitoring writer entities.
 */
#define RTI_MonitoringWriterEntitiesProperty_t_INITIALIZER { \
    NULL, /* publisher_qos */ \
    NULL, /* event_writer_qos */ \
    NULL, /* periodic_writer_qos */ \
    NULL /* logging_writer_qos */ \
}

/*i
 * @brief Finalizes the input monitoring writer entities property.
 */
extern void RTI_MonitoringWriterEntitiesProperty_t_finalize(
        struct RTI_MonitoringWriterEntitiesProperty_t *self);

/*i
 * @brief Create an instance of monitoring writer entities.
 * The writer entities are always created disabled.
 *
 * @param participant \b InOut. The DDS_DomainParticipant used by the
 * monitoring writer entities.
 * @param property \b In. The property for the monitoring writer entities.
 *
 * @return A monitoring writer entities instance or NULL if there is an error.
 */
extern RTI_MonitoringDllExport
struct RTI_MonitoringWriterEntities * RTI_MonitoringWriterEntities_new(
        DDS_DomainParticipant *participant,
        const struct RTI_MonitoringWriterEntitiesProperty_t *property);

/*i
 * @brief Enables the monitoring writer entities.
 * @param self \b InOut. The RTI_MonitoringWriterEntities object.
 * @return RTI_TRUE if the entities were enabled correctly, RTI_FALSE otherwise.
 */
extern RTI_MonitoringDllExport
RTIBool RTI_MonitoringWriterEntities_enable(
        struct RTI_MonitoringWriterEntities *self);

/*i
 * @brief Delete the monitoring writer entities instance.
 * @param self \b In. The monitoring writer entities instance to remove.
 */
extern RTI_MonitoringDllExport 
void RTI_MonitoringWriterEntities_delete(
        struct RTI_MonitoringWriterEntities *self);

  #ifdef __cplusplus
    }	/* extern "C" */
#endif

#endif /* monitoring_monitoringForwarder_h */
/* end of $Id$ */