/*
 * (c) Copyright, Real-Time Innovations, 2022.
 * All rights reserved.
 * No duplications, whole or partial, manual or electronic, may be made
 * without express written permission.  Any such copies, or
 * revisions thereof, must display this notice unaltered.
 * This code contains trade secrets of Real-Time Innovations, Inc.
 */

#ifndef rtitest_connext_interface_h
#define rtitest_connext_interface_h

#include "rtitest/test_dll.h"
#include "rtitest/test_types.h"

#ifdef __cplusplus
extern "C" {
#endif

/************************************ Setup ***********************************/

/** @internal
 * @ingroup RTITestSetting
 * @brief Setup Connext test support.
 */
#define RTITestConnext_setup() \
  /* Setting RTILog_g_enable_backtrace_upon_sigsegv to RTI_TRUE here */ \
  /* because we want to make sure that our tests always enable backtraces. */ \
  RTILog_g_enable_backtrace_upon_sigsegv = RTI_TRUE; \
  /* Setting RTILog_g_detectPrecondition to RTI_TRUE to detect */ \
  /* preconditions by default in tests. */ \
  RTILog_g_detectPrecondition = RTI_TRUE; \
  RTITestConnext_configurePreconditionDetection( \
          &RTILog_g_detectPrecondition, \
          &RTILog_g_preconditionDetected); \
  RTITestConnext_configureAbortOnError(&RTILog_abortOnFatalError)

/******************************** Preconditions *******************************/

/** @internal
 * @ingroup RTITestSetting
 * @brief Configures precondition detection.
 * @param detectPrecondition \b In. Global variable that is set to RTI_TRUE when
 * preconditions want to be detected.
 * @param preconditionDetected \b In. Global variable that is set to RTI_TRUE
 * when a precondition has been detected.
 */
extern RTITestDllExport void RTITestConnext_configurePreconditionDetection(
        RTIBool *detectPrecondition,
        RTIBool *preconditionDetected);

/** @internal
 * @ingroup RTITestSetting
 * @brief Enables or disables the precondition detection configuration.
 * @param enable \b In. If the user passes RTI_TRUE the configuration will be
 * enabled. RTI_FALSE to disable the configuration.
 */
extern RTITestDllExport void RTITestConnext_setPreconditionDetection(
        RTIBool detectPrecondition);

/** @internal
 * @ingroup RTITestSetting
 * @brief Checks if precondition detection is enabled.
 * @return RTI_TRUE if precondition detection is enabled, RTI_FALSE
 * otherwise.
 */
extern RTITestDllExport RTIBool
RTITestConnext_isPreconditionsDetectionEnabled(void);

/** @internal
 * @ingroup RTITestSetting
 * @brief Checks if a precondition has been detected.
 * @return RTI_TRUE if a precondition has been detected, RTI_FALSE otherwise.
 */
extern RTITestDllExport RTIBool RTITestConnext_isPreconditionDetected(void);

/** @internal
 * @ingroup RTITestSetting
 * @brief Resets precondition detection, so
 * RTITestConnext_isPreconditionDetected will return RTI_FALSE again if a
 * precondition has been detected previously.
 */
extern RTITestDllExport void RTITestConnext_resetPreconditionDetection(void);

/** @internal
 * @ingroup RTITestSetting
 * @brief Caches internally if precondition detection is enabled or not.
 */
extern RTITestDllExport void RTITestConnext_cachePreconditionDetection(void);

/** @internal
 * @ingroup RTITestSetting
 * @brief Gets precondition detection cache set by
 * RTITestConnext_cachePreconditionDetection.
 * @return RTI_TRUE if precondition detection was enabled when
 * RTITestConnext_cachePreconditionDetection was called, RTI_FALSE otherwise.
 */
extern RTITestDllExport RTIBool
RTITestConnext_getPreconditionDetectionCache(void);

/******************************* Abort on error *******************************/

/** @internal
 * @ingroup RTITestSetting
 * @brief Configures abort on error.
 * @param abortOnError \b In. Global variable set to RTI_TRUE when abort on
 * error is enabled.
 */
extern RTITestDllExport void RTITestConnext_configureAbortOnError(
        RTIBool *abortOnError);

/** @internal
 * @ingroup RTITestSetting
 * @brief Enables abort on error.
 */
extern RTITestDllExport void RTITestConnext_enableAbortOnError(void);

/** @internal
 * @ingroup RTITestSetting
 * @brief Gets abort on error state.
 * @return RTI_TRUE if abort on error is enabled, RTI_FALSE otherwise.
 */
extern RTITestDllExport RTIBool RTITestConnext_isAbortOnErrorEnabled(void);

/********************* RTI Test log API using Connext log *********************/

/** @internal
 * @ingroup RTITestSetting
 * @brief Log a message using the Connext logging infrastructure instead of the
 * internal one. Expects the message as formatted string and its
 * parameters as variadic arguments.
 * @pre Header "log/log_category.h" should be included to use this macro.
 * @param instrumentationBit__ \b In. RTILogBitmap to define the message's
 * verbosity.
 * @param logMessage__ \b In. Log message template.
 *
 * @details This logging infrastructure is similar to using a printf. If
 * instead you would like a log message to go through the same logging
 * infrastructure as other Connext log messages, use this API. This API is
 * useful, for example, if you have a logging device installed that redirects
 * messages to an output other than stdout.
 */
#define RTITestConnextLogger_log(instrumentationBit__, logMessage__, ...) \
  if (RTITestLog_g_instrumentationMask & (instrumentationBit__)) \
  RTILogMessageParamString_printWithParams( \
          RTI_LOG_PRINT_FORMAT_MASK_ALL, \
          (RTITestLogBitmap) (instrumentationBit__), \
          MODULE_TEST, \
          RTI_FILE_NAME, \
          RTI_LINE_NUMBER, \
          RTI_FUNCTION_NAME, \
          (logMessage__), \
          __VA_ARGS__)

/** @internal
 * @ingroup RTITestSetting
 * @brief Log a message with a custom function name using Connext logging
 * infrastructure instead of the internal one. Expects the
 * message as format string and its parameters as variadic arguments.
 * @pre Header "log/log_category.h" should be included to use this macro.
 * @param formatMask__ \b In. RTILogBitmap to define the message's format mask.
 * @param instrumentationBit__ \b In. RTILogBitmap to define the message's
 * verbosity.
 * @param functionName__ \b In. Name of the function that is logging this
 * message.
 *
 * @details Each log message output by Connext includes the method name of the
 * function that is logging the message. If you would like to include a
 * different function name than the one that is actually logging the message,
 * use this API. This API is useful, for example, if you are using a small
 * helper function to log a message and you would like the caller's function
 * name to be logged instead of the helper function.
 */
#define RTITestConnextLogger_logWithFunctionName( \
        formatMask__, \
        instrumentationBit__, \
        functionName__, \
        ...) \
  if (RTITestLog_g_instrumentationMask & (instrumentationBit__)) \
  RTILogMessage_printWithParams( \
          (formatMask__) \
                  & (RTILogPrintFormatKindMask) ~( \
                          RTI_LOG_PRINT_BIT_LOG_LEVEL), \
          (RTITestLogBitmap) (instrumentationBit__), \
          MODULE_TEST, \
          RTI_FILE_NAME, \
          RTI_LINE_NUMBER, \
          (functionName__), \
          __VA_ARGS__)

/** @internal
 * @ingroup RTITestSetting
 * @brief Log a message with free form using Connext logging infrastructure
 * instead of the internal one. Expects the message as format string and its
 * parameters as variadic arguments.
 * @pre Header "log/log_common.h" should be included to use this macro.
 * @param instrumentationBit__ \b In. RTILogBitmap to define the message's
 * verbosity.
 */
#define RTITestConnextLogger_logParamString(instrumentationBit__, ...) \
  if ((RTITestLog_g_instrumentationMask) & (instrumentationBit__)) \
  RTILogParamString_printUnformatted( \
          (RTITestLogBitmap) (instrumentationBit__), \
          __VA_ARGS__)

#ifdef __cplusplus
}
#endif

#endif /* rtitest_connext_interface_h */