/**
 * (c) Copyright, Real-Time Innovations, 2021
 * All rights reserved.
 * No duplications, whole or partial, manual or electronic, may be made
 * without express written permission.  Any such copies, or
 * revisions thereof, must display this notice unaltered.
 * This code contains trade secrets of Real-Time Innovations, Inc.
 *
 * Description
 * -----------
 * RTI definitions of cryptography library structures and macros to enable
 * supporting for WolfSSL.
 * A user application does not need to include this header file. It's only
 * intended for internal use by other connextdds modules.
 * =========================================================================
 */

#ifndef security_crypto_lib_adapter_wolfssl_h
#define security_crypto_lib_adapter_wolfssl_h

#define RTI_WOLFSSL_VERSION_4_7_0 0x04007000
#define RTI_WOLFSSL_VERSION_5_5_0 0x05005000

#ifdef RTI_WOLFSSL

#include <wolfcrypt/coding.h>  /* for Base64_Decode */
#include <wolfssl/version.h>

#ifdef RTI_SECURITY_PLUGINS_SUPPORT_THREAD_SPECIFIC_STORAGE
  #undef RTI_SECURITY_PLUGINS_SUPPORT_THREAD_SPECIFIC_STORAGE
#endif

#define RTI_SUPPORT_X509_V3_KEY_USAGE_EXTENSION
#define RTI_SUPPORT_X509_V3_KEY_USAGE_EXTENSION_ENFORCEMENT
#define X509v3_KU_KEY_CERT_SIGN KU_KEY_CERT_SIGN

/*
 * Symbols to close the gap between OpenSSL and WolfSSL when needed.
 */
#ifndef PKCS5_DEFAULT_ITER
  #define PKCS5_DEFAULT_ITER 2048
#endif

#ifndef LN_sha256
  #define LN_sha256 "sha256"
#endif

/*
 * OpenSSL returns 1 if two keys match (and they are the same type).
 * wolfSSL behaves the same only if it was compiled with
 * WOLFSSL_ERROR_CODE_OPENSSL. By default, wolfSSL returns 0 when two keys
 * match. This adapter macro takes care of handling the different return value
 * with respect to OpenSSL.
 */
#if defined(WOLFSSL_ERROR_CODE_OPENSSL)
  #define RTI_Security_CryptoLibAdapterEvpPkey_equals(pkey1__, pkey2__) \
      (EVP_PKEY_cmp(pkey1__, pkey2__))
#else
  #define RTI_Security_CryptoLibAdapterEvpPkey_equals(pkey1__, pkey2__) \
      (EVP_PKEY_cmp(pkey1__, pkey2__) == 0)
#endif

#ifdef __cplusplus
extern "C" {
#endif

#define WOLF_STACK_OF_X509_CRL STACK_OF(X509_INFO)

WOLF_STACK_OF_X509_CRL * RTI_Security_CryptoLibAdapterWolfSSL_sk_X509_CRL_new_null(void);
#define sk_X509_CRL_new_null RTI_Security_CryptoLibAdapterWolfSSL_sk_X509_CRL_new_null

typedef void (*stack_free_func)(void *);

typedef int (*X509_STORE_CTX_verify_fn)(WOLFSSL_X509_STORE_CTX *);

extern
void RTI_Security_CryptoLibAdapterWolfSSL_logging_cb(
        const int logLevel,
        const char *const logMessage);
#define wolfSSL_logging_cb RTI_Security_CryptoLibAdapterWolfSSL_logging_cb

extern
int RTI_Security_CryptoLibAdapterWolfSSL_verify_cb_dummy(
        int cb_value,
        X509_STORE_CTX *store_ctx);

extern int RTI_Security_CryptoLibAdapterWolfSSL_EVP_PKEY_set0(
        WOLFSSL_EVP_PKEY *pkey,
        unsigned char *data,
        int dataLength);
#define EVP_PKEY_set0 RTI_Security_CryptoLibAdapterWolfSSL_EVP_PKEY_set0

#define X509_STORE_set_verify_func X509_STORE_set_verify_cb

extern
WOLFSSL_X509 * RTI_Security_CryptoLibAdapterWolfSSL_PEM_read_bio_X509_AUX(
        WOLFSSL_BIO *bp,
        WOLFSSL_X509 **x,
        pem_password_cb *cb,
        void *u);
#define RTI_Security_CryptoLibAdapter_PEM_read_bio_X509_AUX \
    RTI_Security_CryptoLibAdapterWolfSSL_PEM_read_bio_X509_AUX

extern
WOLFSSL_BIO * RTI_Security_CryptoLibAdapterWolfSSL_BIO_new_mem_buf(
        const void *buf,
        int len);
#define RTI_Security_CryptoLibAdapter_BIO_new_mem_buf \
    RTI_Security_CryptoLibAdapterWolfSSL_BIO_new_mem_buf

extern
int RTI_Security_CryptoLibAdapterWolfSSL_X509_NAME_print_ex(
        WOLFSSL_BIO *bio,
        WOLFSSL_X509_NAME *name,
        int indent,
        unsigned long flags);
#define RTI_Security_CryptoLibAdapter_X509_NAME_print_ex \
    RTI_Security_CryptoLibAdapterWolfSSL_X509_NAME_print_ex

/*
 * This operation resets a BIO to its initial state, before anything was read
 * from it. Returns DDS_BOOLEAN_TRUE on success. See
 * https://www.wolfssl.com/doxygen/group__IO.html#ga87c06554b6c8af8870bee7e4a0b6165a
 */
#define RTI_Security_CryptoLibAdapter_BIO_reset(bio__) \
  (BIO_reset(bio__) == 0)

extern
RTIBool RTI_Security_CryptoLibAdapterWolfSSL_verifyPKCS7(
        const PKCS7 *p7,
        WOLFSSL_STACK *mainAuthority,
        WOLFSSL_STACK *const *altAuthorities,
        unsigned int altAuthoritiesCount);

#ifdef  __cplusplus
}
#endif

/***************************************************************************
 * EVP_CIPHER_CTX
 ***************************************************************************/
  #define RTI_Security_CryptoLibAdapterEvpCipherCtxHandle EVP_CIPHER_CTX

  #define RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_getCryptoLibCtx(ctxHandle__) \
      (ctxHandle__)

  #define RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_initialize( \
          ctxHandle__, \
          cryptoContext__) \
      (EVP_CIPHER_CTX_init(ctxHandle__), (1))

  #define RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_reset(ctxHandle__) \
      EVP_CIPHER_CTX_cleanup(ctxHandle__); \
      EVP_CIPHER_CTX_init(ctxHandle__);

  #define RTI_Security_CryptoLibAdapterEvpCipherCtxHandle_finalize( \
          ctxHandle__, \
          cryptoContext__) \
      RTI_Security_Utility_unusedParameter(cryptoContext__); \
      EVP_CIPHER_CTX_cleanup(ctxHandle__)

/******************************************************************************
 * EVP_MD_CTX
 ******************************************************************************/
#define RTI_Security_CryptoLibAdapterEvpMdCtxHandle WOLFSSL_EVP_MD_CTX
#define RTI_Security_CryptoLibAdapterEvpMdCtxHandle_getCryptoLibCtx(ctxHandle__) (ctxHandle__)
#define RTI_Security_CryptoLibAdapterEvpMdCtxHandle_initialize( \
          ctxHandle__, \
          cryptoContext__) \
      (EVP_MD_CTX_init(ctxHandle__), (1))
#define RTI_Security_CryptoLibAdapterEvpMdCtxHandle_finalize( \
          ctxHandle__, \
          cryptoContext__) \
      RTI_Security_Utility_unusedParameter(cryptoContext__); \
      EVP_MD_CTX_cleanup(ctxHandle__)
#define RTI_Security_CryptoLibAdapterEvpMdCtxHandle_reset(ctxHandle__) \
      EVP_MD_CTX_cleanup(ctxHandle__); \
      EVP_MD_CTX_init(ctxHandle__);

/* Note: We are not using EVP_ENCODE_CTX in code for WolfSSL. */

/***************************************************************************
 * EVP_PKEY_new_mac_key
 **************************************************************************/
#define RTI_Security_CryptoLibAdapterEvpNewMacKey EVP_PKEY_new_mac_key

/***************************************************************************
 * Signature Identifiers
 **************************************************************************/

  /*
   * wolfSSL doesn't support RSA PSS certificates, so it doesn't have this
   * signature identifier.
   * However, wolfSSL supports signing with a RSA PKCS key using PSS padding.
   * This identifier represents this scenario. As wolfSSL doesn't provide the
   * identifier itself, we are using a custom one for our implementation.
   *
   * Therefore RTI_SECURITY_SIGNATURE_ID_RSA_PSS_SHA256 represents the signature
   * from RSA keys (either PSS or PKCS) using PSS padding.
   */
  #define RTI_SECURITY_SIGNATURE_ID_RSA_PSS_SHA256 404

  /*
   * Unknown Curve Identifier.
   * This value is crypto-library dependent because wolfSSL differentiates
   * between NIDs that are used for ASN objects and the ones used in EC curves.
   * Each type has its own enum, and mixing them (NID_undef is an identifier
   * for ASN objects, not EC curves) generates static analysis warnings.
   * Therefore, our unknown value in wolfSSL is going to be a curve that we
   * don't support.
   */
  #define RTI_SECURITY_EC_CURVE_ID_UNKNOWN POINT_CONVERSION_COMPRESSED

/*
 * Prevent a warning due to EVP_DigestSignUpdate accepting a size_t count
 * in OpenSSL, but an unsigned int count in wolfSSL.
 */
typedef unsigned int RTI_Security_CryptoLibAdapterDigestSignUpdateCount;


/*
  * wolfSSL 4.7 does not implement X509_STORE_CTX_get0_chain, so we create
  * a temporal adapter for it.
  * Future wolfSSL versions do support X509_STORE_CTX_get0_chain, and the
  * adapter may not be necessary in the future.
  */
#define RTI_Security_CryptoLibAdapterX509StoreCtxGetChain \
    wolfSSL_X509_STORE_CTX_get_chain

#if (LIBWOLFSSL_VERSION_HEX < RTI_WOLFSSL_VERSION_4_7_0)
  #error "wolfSSL library version error: " \
        "version < 4.7.0 is not supported"
#elif (LIBWOLFSSL_VERSION_HEX == RTI_WOLFSSL_VERSION_4_7_0)
  #include "security_crypto_lib_adapter_wolfssl47.h"
#elif (LIBWOLFSSL_VERSION_HEX >= RTI_WOLFSSL_VERSION_5_5_0)
  #include "security_crypto_lib_adapter_wolfssl55.h"
#else
  #error "wolfSSL library version error: " \
        "4.7.0 < version < 5.5.0 is not supported"
#endif /* LIBWOLFSSL_VERSION_HEX */

#endif /* RTI_WOLFSSL */
#endif /* security_crypto_lib_adapter_wolfssl_h */
